"""Image functions for PET data reconstruction and processing."""
__author__      = "Pawel J. Markiewicz"
__copyright__   = "Copyright 2018"
#-------------------------------------------------------------------------------

import sys
import os

import numpy as np
import math
from math import pi
import random
import scipy.ndimage as ndi

import nibabel as nib
import pydicom as dcm
import re
import glob

from subprocess import call
import time
import multiprocessing

from niftypet.nipet import mmraux
from niftypet import nimpa

import resources as rs


#===================================================================================
# IMAGE ROUTINES
#===================================================================================

def convert2e7(img, Cnt):
    '''Convert GPU optimised image to Siemens/E7 image shape (127,344,344).'''

    margin = (Cnt['SO_IMX']-Cnt['SZ_IMX'])/2

    #permute the dims first
    imo = np.transpose(img, (2,0,1))

    nvz = img.shape[2]

    #> get the x-axis filler and apply it
    filler = np.zeros((nvz, Cnt['SZ_IMY'], margin), dtype=np.float32)
    imo = np.concatenate((filler, imo, filler), axis=2)

    #> get the y-axis filler and apply it
    filler = np.zeros((nvz, margin, Cnt['SO_IMX']), dtype=np.float32)
    imo = np.concatenate((filler, imo, filler), axis=1)
    return imo

def convert2dev(im, Cnt):
    '''Reshape Siemens/E7 (default) image for optimal GPU execution.'''

    if im.shape[1]!=Cnt['SO_IMY'] or im.shape[2]!=Cnt['SO_IMX']:
        raise ValueError('e> input image array is not of the correct Siemens shape.')
    
    if 'rSZ_IMZ' in Cnt and im.shape[0]!=Cnt['rSZ_IMZ']:
        print 'w> the axial number of voxels does not match the reduced rings.'
    elif not 'rSZ_IMZ' in Cnt and im.shape[0]!=Cnt['SZ_IMZ']:
        print 'w> the axial number of voxels does not match the rings.'

    im_sqzd = np.zeros((im.shape[0], Cnt['SZ_IMY'], Cnt['SZ_IMX']), dtype=np.float32)
    margin = (Cnt['SO_IMX']-Cnt['SZ_IMX'])/2
    margin_=-margin
    if margin==0: 
        margin = None
        margin_= None

    im_sqzd = im[:, margin:margin_, margin:margin_]
    im_sqzd = np.transpose(im_sqzd, (1, 2, 0))
    return im_sqzd

def cropxy(im, imsize, datain, Cnt, store_pth=''):
    '''Crop image transaxially to the size in tuple <imsize>.  
    Return the image and the affine matrix.
    '''
    if not imsize[0]%2==0 and not imsize[1]%2==0:
        print 'e> image size has to be an even number!'
        return None

    # cropping indexes
    i0 = (Cnt['SO_IMX']-imsize[0])/2
    i1 = (Cnt['SO_IMY']+imsize[1])/2

    B = image_affine(datain, Cnt, gantry_offset=False)
    B[0,3] -= 10*Cnt['SO_VXX']*i0
    B[1,3] += 10*Cnt['SO_VXY']*(Cnt['SO_IMY']-i1)

    cim = im[:, i0:i1, i0:i1]

    if store_pth!='':
        nimpa.array2nii( cim[::-1,::-1,:], B, store_pth, descrip='cropped')
        if Cnt['VERBOSE']:  print 'i> saved cropped image to:', store_pth

    return cim, B
#-------------------------------------------------------------------------------------------

def image_affine(datain, Cnt, gantry_offset=False):
    '''Creates a blank reference image, to which another image will be resampled'''

    #------get necessary data for -----
    # gantry offset
    if gantry_offset:
        goff, tpo = mmraux.lm_pos(datain, Cnt)
    else:
        goff = np.zeros((3))
    vbed, hbed = mmraux.vh_bedpos(datain, Cnt)
    # create a reference empty mu-map image
    B = np.diag(np.array([-10*Cnt['SO_VXX'], 10*Cnt['SO_VXY'], 10*Cnt['SO_VXZ'], 1]))
    B[0,3] = 10*(.5*Cnt['SO_IMX']*Cnt['SO_VXX']      + goff[0])
    B[1,3] = 10*((-.5*Cnt['SO_IMY']+1)*Cnt['SO_VXY'] - goff[1])
    B[2,3] = 10*((-.5*Cnt['SO_IMZ']+1)*Cnt['SO_VXZ'] - goff[2] + hbed)
    # -------------------------------------------------------------------------------------
    return B


#================================================================================

def getmu_off(mu, Cnt, Offst=np.array([0., 0., 0.])):
    #number of voxels
    nvx = mu.shape[0]
    #change the shape to 3D
    mu.shape = (Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX'])

    #-------------------------------------------------------------------------
    # CORRECT THE MU-MAP for GANTRY OFFSET
    #-------------------------------------------------------------------------
    Cim = {
        'VXSOx':0.208626,
        'VXSOy':0.208626,
        'VXSOz':0.203125,
        'VXNOx':344,
        'VXNOy':344,
        'VXNOz':127,

        'VXSRx':0.208626,
        'VXSRy':0.208626,
        'VXSRz':0.203125,
        'VXNRx':344,
        'VXNRy':344,
        'VXNRz':127
    }
    #original image offset
    Cim['OFFOx'] = -0.5*Cim['VXNOx']*Cim['VXSOx']
    Cim['OFFOy'] = -0.5*Cim['VXNOy']*Cim['VXSOy']
    Cim['OFFOz'] = -0.5*Cim['VXNOz']*Cim['VXSOz']
    #resampled image offset
    Cim['OFFRx'] = -0.5*Cim['VXNRx']*Cim['VXSRx']
    Cim['OFFRy'] = -0.5*Cim['VXNRy']*Cim['VXSRy']
    Cim['OFFRz'] = -0.5*Cim['VXNRz']*Cim['VXSRz']
    #transformation matrix
    A = np.array(
        [[ 1., 0., 0.,  Offst[0] ],
        [  0., 1., 0.,  Offst[1] ],
        [  0., 0., 1.,  Offst[2] ],
        [  0., 0., 0.,  1. ]], dtype=np.float32
        )
    #apply the gantry offset to the mu-map
    mur = nimpa.prc.improc.resample(mu, A, Cim)
    return mur

def getinterfile_off(fmu, Cnt, Offst=np.array([0., 0., 0.])):
    '''
    Return the floating point mu-map in an array from Interfile, accounting for image offset (does slow interpolation).
    '''
    #read the image file
    f = open(fmu, 'rb')
    mu = np.fromfile(f, np.float32)
    f.close()
    
    # save_im(mur, Cnt, os.path.dirname(fmu) + '/mur.nii')
    #-------------------------------------------------------------------------
    mur = getmu_off(mu, Cnt)
    #create GPU version of the mu-map
    murs = convert2dev(mur, Cnt)
    #get the basic stats
    mumax = np.max(mur)
    mumin = np.min(mur)
    #number of voxels greater than 10% of max image value
    n10mx = np.sum(mur>0.1*mumax)
    #return image dictionary with the image itself and some other stats
    mu_dct = {'im':mur,
              'ims':murs,
              'max':mumax,
              'min':mumin,
              'nvx':nvx,
              'n10mx':n10mx}
    return mu_dct

#================================================================================
def getinterfile(fim, Cnt):
    '''Return the floating point image file in an array from an Interfile file.'''
    
    #read the image file
    f = open(fim, 'rb')
    im = np.fromfile(f, np.float32)
    f.close()

    #number of voxels
    nvx = im.shape[0]
    #change the shape to 3D
    im.shape = (Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX'])

    #get the basic stats
    immax = np.max(im)
    immin = np.min(im)

    #number of voxels greater than 10% of max image value
    n10mx = np.sum(im>0.1*immax)

    #reorganise the image for optimal gpu execution
    im_sqzd = convert2dev(im, Cnt)

    #return image dictionary with the image itself and some other stats
    im_dct = {'im':im,
              'ims':im_sqzd,
              'max':immax,
              'min':immin,
              'nvx':nvx,
              'n10mx':n10mx}

    return im_dct
#================================================================================


#=================================================================================
#-define uniform cylinder
def get_cylinder(Cnt, rad=25, xo=0, yo=0, unival=1, gpu_dim=False):
    '''Outputs image with a uniform cylinder of intensity = unival, radius = rad, and transaxial centre (xo, yo)'''
    imdsk = np.zeros((1, Cnt['SO_IMX'], Cnt['SO_IMY']), dtype=np.float32)
    for t in np.arange(0, math.pi, math.pi/(2*360)):
        x = xo+rad*math.cos(t)
        y = yo+rad*math.sin(t)
        yf = np.arange(-y+2*yo, y, Cnt['SO_VXY']/2)
        v = np.int32(.5*Cnt['SO_IMX'] - np.ceil(yf/Cnt['SO_VXY']))
        u = np.int32(.5*Cnt['SO_IMY'] + np.floor(x/Cnt['SO_VXY']))
        imdsk[0,v,u] = unival
    if 'rSO_IMZ' in Cnt:
        nvz = Cnt['rSO_IMZ']
    else:
        nvz = Cnt['SO_IMZ']
    imdsk = np.repeat(imdsk, nvz, axis=0)
    if gpu_dim: imdsk = convert2dev(imdsk, Cnt)
    return imdsk


#================================================================================
def hu2mu(im):
    '''HU units to 511keV PET mu-values'''

    # convert nans to -1024 for the HU values only
    im[np.isnan(im)] = -1024
    # constants
    muwater  = 0.096
    mubone   = 0.172
    rhowater = 0.158
    rhobone  = 0.326
    uim = np.zeros(im.shape, dtype=np.float32)
    uim[im<=0] = muwater * ( 1+im[im<=0]*1e-3 )
    uim[im> 0] = muwater * \
        ( 1+im[im>0]*1e-3 * rhowater/muwater*(mubone-muwater)/(rhobone-rhowater) )
    # remove negative values
    uim[uim<0] = 0
    return uim


# =====================================================================================
# object/patient mu-map resampling to NIfTI
# better use dcm2niix
def mudcm2nii(datain, Cnt):
    '''DICOM mu-map to NIfTI'''

    mu, pos, ornt = nimpa.dcm2im(datain['mumapDCM'])
    mu *= 0.0001
    A = pos['AFFINE']
    A[0,0] *= -1
    A[0,3] *= -1
    A[1,3] += A[1,1]
    nimpa.array2nii(mu[:,::-1,:], A, os.path.join(os.path.dirname(datain['mumapDCM']),'mu.nii.gz'))

    #------get necessary data for creating a blank reference image (to which resample)-----
    # gantry offset
    goff, tpo = mmraux.lm_pos(datain, Cnt)
    ihdr, csainfo = mmraux.hdr_lm(datain)
    #start horizontal bed position
    p = re.compile(r'start horizontal bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    hbedpos = 0.1*float(ihdr[m.start()+fi+1:m.end()])

    B = np.diag(np.array([-10*Cnt['SO_VXX'], 10*Cnt['SO_VXY'], 10*Cnt['SO_VXZ'], 1]))
    B[0,3] = 10*(.5*Cnt['SO_IMX']*Cnt['SO_VXX']      + goff[0])
    B[1,3] = 10*((-.5*Cnt['SO_IMY']+1)*Cnt['SO_VXY'] - goff[1])
    B[2,3] = 10*((-.5*Cnt['SO_IMZ']+1)*Cnt['SO_VXZ'] - goff[2] + hbedpos)
    im = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)
    nimpa.array2nii(im, B, os.path.join(os.path.dirname(datain['mumapDCM']),'muref.nii.gz'))
    # -------------------------------------------------------------------------------------
    fmu = os.path.join(os.path.dirname(datain['mumapDCM']),'mu_r.nii.gz')
    if os.path.isfile( Cnt['RESPATH'] ):
        call( [ Cnt['RESPATH'],
                    '-ref', os.path.join(os.path.dirname(datain['mumapDCM']),'muref.nii.gz'),
                    '-flo', os.path.join(os.path.dirname(datain['mumapDCM']),'mu.nii.gz'),
                    '-res', fmu,
                    '-pad', '0'] )
    else:
        print 'e> path to resampling executable is incorrect!'
        raise IOError('Error launching NiftyReg for image resampling.')

    return fmu

# =====================================================================================
def obj_mumap(
        datain,
        params={},
        outpath='',
        comment='',
        store=False,
        store_npy=False,
        gantry_offset=True,
        del_auxilary=True,
        ):
    '''Get the object mu-map from DICOM images'''

    
    # three ways of passing scanner constants <Cnt> are here decoded
    if 'Cnt' in params:
        Cnt = params['Cnt']
    elif 'SO_IMZ' in params:
        Cnt = params
    else:
        Cnt = rs.get_mmr_constants()


    # output folder
    if outpath=='':
        fmudir = os.path.join( datain['corepath'], 'mumap-obj' )
    else:
        fmudir = os.path.join( outpath, 'mumap-obj' )
    nimpa.create_dir(fmudir)
    # ref file name
    fmuref = os.path.join(fmudir, 'muref.nii.gz')
    # ref affine
    B = image_affine(datain, Cnt, gantry_offset=gantry_offset)
    # ref image (blank)
    im = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)
    # store ref image
    nimpa.array2nii(im, B, fmuref)

    # check if the object dicom files for MR-based mu-map exists
    if not 'mumapDCM' in datain or not os.path.isdir(datain['mumapDCM']):
        print 'e> DICOM forlder for the mu-map does not exist.'
        return None

    fnii = 'converted-from-object-DICOM_'
    tstmp = nimpa.time_stamp(simple_ascii=True)

    # find residual(s) from previous runs and delete them
    resdcm = glob.glob( os.path.join(fmudir, '*'+fnii+'*.nii*') )
    for d in resdcm:
        os.remove(d)

    # convert the DICOM mu-map images to nii
    call( [ Cnt['DCM2NIIX'], '-f', fnii+tstmp, '-o', fmudir, datain['mumapDCM'] ] )
    #files for the T1w, pick one:
    fmunii = glob.glob( os.path.join(fmudir, '*'+fnii+tstmp+'*.nii*') )[0]
    # fmunii = glob.glob( os.path.join(datain['mumapDCM'], '*converted*.nii*') )
    # fmunii = fmunii[0]

    # the converted nii image resample to the reference size
    fmu = os.path.join(fmudir, comment+'mumap_tmp.nii.gz')
    if os.path.isfile( Cnt['RESPATH'] ):
        cmd = [ Cnt['RESPATH'],
                    '-ref', fmuref,
                    '-flo', fmunii,
                    '-res', fmu,
                    '-pad', '0']
        if not Cnt['VERBOSE']: cmd.append('-voff')
        call(cmd)
    else:
        print 'e> path to resampling executable is incorrect!'
        sys.exit()

    nim = nib.load(fmu)
    # get the affine transform
    A = nim.get_sform()
    mu = nim.get_data()
    mu = np.transpose(mu[:,::-1,::-1], (2, 1, 0))
    # convert to mu-values
    mu = np.float32(mu)/1e4
    mu[mu<0] = 0
    
    #> return image dictionary with the image itself and some other stats
    mu_dct = {  'im':mu,
                'affine':A}
    if not del_auxilary: mu_dct['fmuref'] = fmuref

    # store the mu-map if requested
    if store_npy:
        # to numpy array
        fnp = os.path.join(fmudir, 'mumap-from-DICOM.npy' )
        np.save(fnp, (mu, A))
    
    if store:
        # with this file name
        fmumap = os.path.join(fmudir, 'mumap-from-DICOM_no-alignment'+comment+'.nii.gz')
        nimpa.array2nii(mu[::-1,::-1,:], A, fmumap)
        mu_dct['fim'] = fmumap

    if del_auxilary:
        os.remove(fmuref)
        os.remove(fmunii)
        os.remove(fmu)


    return mu_dct

#=================================================================================
# pCT/UTE MU-MAP ALIGNED
#---------------------------------------------------------------------------------
def align_mumap(
        datain,
        scanner_params={},
        outpath='',
        use_stored=False,
        hst=[],
        t0=0, t1=0,
        itr=2,
        faff='',
        fpet='',
        fcomment='',
        store=False,
        store_npy=False,
        petopt='ac',
        musrc='ute', # another option is pct for mu-map source
        ute_name='UTE2',
        del_auxilary=True,
        verbose=True,
    ):
    '''
    Align the a pCT or MR-derived mu-map to a PET image reconstructed to chosen
    specifications (e.g., with/without attenuation and scatter corrections)

    use_sotred only works if hst or t0/t1 given but not when faff.
    ''' 


    #> output folder
    if outpath=='':
        opth = os.path.join( datain['corepath'], 'mumap-obj' )
    else:
        opth = os.path.join( outpath, 'mumap-obj' )
    
    #> create the folder, if not existent
    nimpa.create_dir(opth)

    #> get the timing of PET if affine not given
    if faff!='' and 't0' in hst:
        t0 = hst['t0']
        t1 = hst['t1']

    #> file name for the output mu-map
    fnm = 'mumap-' + musrc.upper()

    #> output dictionary
    mu_dct = {}

    #---------------------------------------------------------------------------
    #> used stored if requested
    if use_stored and affine!='':
        fmu_stored = fnm + '-aligned-to_t'\
                     + str(hst['t0'])+'-'+str(hst['t1'])+'_'+petopt.upper()\
                     + fcomment
        fmupath = os.path.join(opth, fmu_stored)
        if os.path.isfile( fmupath ):
            mudct_stored = nimpa.getnii(fmupath, output='all')
            #> create output dictionary
            mu_dct['im'] = mudct_stored['im']
            mu_dct['affine'] = mudct_stored['affine']
            #mu_dct['faff'] = faff
            return mu_dct
    #---------------------------------------------------------------------------

    #> three ways of passing scanner constants <Cnt> are here decoded
    if 'Cnt' in scanner_params:
        Cnt = scanner_params['Cnt']
    elif 'SO_IMZ' in scanner_params:
        Cnt = scanner_params
    else:
        Cnt = rs.get_mmr_constants()
    Cnt   = scanner_params['Cnt']

    #> if affine not provided histogram the LM data for recon and registration
    if not os.path.isfile(faff):
        from niftypet.nipet.prj import mmrrec
        #-histogram the list data if needed
        if not hst:
            from niftypet.nipet import mmrhist
            if 'txLUT' in scanner_params:
                hst = mmrhist(datain, scanner_params, t0=t0, t1=t1)
            else:
                raise ValueError(
                    'e> full scanner parameters not provided\
                     but required for histogramming.')

    #=========================================================
    #-get hardware mu-map
    if 'hmumap' in datain and os.path.isfile(datain['hmumap']):
        muh, _, _ = np.load(datain['hmumap'])
        if verbose: print 'i> loaded hardware mu-map from file:', datain['hmumap']
    elif outpath!='':
        hmupath = os.path.join( os.path.join(outpath,'mumap-hdw'), 'hmumap.npy')
        if os.path.isfile( hmupath ):
            muh, _, _ = np.load(hmupath)
            datain['hmumap'] = hmupath
        else:
            raise IOError('Invalid path to the hardware mu-map')
    else:
        print 'e> obtain the hardware mu-map first.'
        raise IOError('Could not find the hardware mu-map.  Have you run the routine for hardware mu-map?')
    #=========================================================
    #-check if T1w image is available
    if not 'MRT1W#' in datain and not 'T1nii' in datain and not 'T1bc' in datain \
    and not 'T1N4' in datain:
        print 'e> no MR T1w images required for co-registration!'
        raise IOError('T1w image could not be obtained')
    #=========================================================

    #-if the affine is not given, 
    #-it will be generated by reconstructing PET image, with some or no corrections
    if not os.path.isfile(faff):
        # first recon pet to get the T1 aligned to it
        if petopt=='qnt':
            # ---------------------------------------------
            # OPTION 1 (quantitative recon with all corrections using MR-based mu-map)
            # get UTE object mu-map (may not be in register with the PET data)
            mudic = obj_mumap(
                        datain,
                        Cnt,
                        del_auxilary=del_auxilary)
            muo = mudic['im']
            # reconstruct PET image with UTE mu-map to which co-register T1w
            recout = mmrrec.osemone(   
                datain, [muh, muo], 
                hst, scanner_params,
                recmod=3, itr=itr, fwhm=0.,
                fcomment=fcomment+'_QNT-UTE',
                outpath=os.path.join(outpath, 'PET', 'positioning'),
                store_img=True)
        elif petopt=='nac':
            # ---------------------------------------------
            # OPTION 2 (recon without any corrections for scatter and attenuation)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            muo = np.zeros(muh.shape, dtype=muh.dtype)
            recout = mmrrec.osemone(
                datain, [muh, muo],
                hst, scanner_params, 
                recmod=1, itr=itr, fwhm=0., 
                fcomment=fcomment+'_NAC',
                outpath=os.path.join(outpath,'PET', 'positioning'),
                store_img=True)
        elif petopt=='ac':
            # ---------------------------------------------
            # OPTION 3 (recon with attenuation correction only but no scatter)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            mudic = obj_mumap(
                    datain,
                    Cnt,
                    del_auxilary=del_auxilary)
            muo = mudic['im']

            recout = mmrrec.osemone(
                datain, [muh, muo],
                hst, scanner_params, 
                recmod=1, itr=itr, fwhm=0., 
                fcomment=fcomment+'_AC-UTE',
                outpath=os.path.join(outpath,'PET', 'positioning'),
                store_img=True)       

        fpet = recout.fpet
        mu_dct['fpet'] = fpet

        #------------------------------
        if musrc=='ute' and ute_name in datain and os.path.exists(datain[ute_name]):
            # change to NIfTI if the UTE sequence is in DICOM files (folder)
            if os.path.isdir(datain[ute_name]):
                fnew =  os.path.basename(datain[ute_name])
                call( [ Cnt['DCM2NIIX'], '-f', fnew, datain[ute_name] ] )
                fute = glob.glob(os.path.join(datain[ute_name], fnew+'*nii*'))[0]
            elif os.path.isfile(datain[ute_name]):
                fute = datain[ute_name]
            
            # get the affine transformation
            try:
                regdct = nimpa.coreg_spm(
                    fpet,
                    fute,
                    outpath=os.path.join(outpath,'PET', 'positioning')
                )
            except:
                regdct = nimpa.affine_niftyreg(
                    fpet,
                    fute,
                    outpath=os.path.join(outpath,'PET', 'positioning'),
                    #fcomment=fcomment,
                    executable = Cnt['REGPATH'],
                    omp = multiprocessing.cpu_count()/2,
                    rigOnly = True,
                    affDirect = False,
                    maxit=5,
                    speed=True,
                    pi=50, pv=50,
                    smof=0, smor=0,
                    rmsk=True,
                    fmsk=True,
                    rfwhm=15., #millilitres
                    rthrsh=0.05,
                    ffwhm = 15., #millilitres
                    fthrsh=0.05,
                    verbose=verbose
                )
            
            faff_mrpet = regdct['faff']
            
        elif musrc=='pct':

            ft1w = nimpa.pick_t1w(datain)

            try:
                regdct = nimpa.coreg_spm(
                    fpet,
                    ft1w,
                    outpath=os.path.join(outpath,'PET', 'positioning')
                )
            except:
                regdct = nimpa.affine_niftyreg(
                    fpet,
                    ft1w,
                    outpath=os.path.join(outpath,'PET', 'positioning'),
                    executable = Cnt['REGPATH'],
                    omp = multiprocessing.cpu_count()/2,
                    rigOnly = True,
                    affDirect = False,
                    maxit=5,
                    speed=True,
                    pi=50, pv=50,
                    smof=0, smor=0,
                    rmsk=True,
                    fmsk=True,
                    rfwhm=15., #millilitres
                    rthrsh=0.05,
                    ffwhm = 15., #millilitres
                    fthrsh=0.05,
                    verbose=verbose
                )

            faff_mrpet = regdct['faff']

        else: 
            raise IOError('Floating MR image not provided or is invalid.')

    else:
        faff_mrpet = faff
        regdct = {}
        if not os.path.isfile(fpet):
            raise IOError('e> the reference PET should be supplied with the affine.')
        
    #> output file name for the aligned mu-maps
    if musrc=='pct':
        freg = os.path.join(opth, 'pCT-res-tmp'+fcomment+'.nii.gz')
        fflo = datain['pCT']

    elif musrc=='ute':
        freg = os.path.join(opth, 'UTE-res-tmp'+fcomment+'.nii.gz')
        if 'UTE' not in datain:
            fnii = 'converted-from-DICOM_'
            tstmp = nimpa.time_stamp(simple_ascii=True)
            # convert the DICOM mu-map images to nii
            if 'mumapDCM' not in datain: 
                raise IOError('DICOM with the UTE mu-map are not given.')
            call( [ Cnt['DCM2NIIX'], '-f', fnii+tstmp, '-o', opth, datain['mumapDCM'] ] )
            #files for the T1w, pick one:
            fflo = glob.glob( os.path.join(opth, '*'+fnii+tstmp+'*.nii*') )[0]
        else:
            if os.path.isfile(datain['UTE']):
                fflo = datain['UTE']
            else:
                raise IOError('The provided NIfTI UTE path is not valid.')

    #> call the resampling routine to get the pCT/UTE in place
    if 'matlab_eng' in regdct:
        nimpa.resample_spm(
            fpet,
            fflo,
            faff_mrpet,
            matlab_eng=regdct['matlab_eng'],
            fimout=freg,
            del_ref_uncmpr=True,
            del_flo_uncmpr=True,
            del_out_uncmpr=True
        )
    else:
        nimpa.resample_niftyreg(
            fpet,
            fflo,
            faff_mrpet,
            fimout = freg,
            executable = Cnt['RESPATH'],
            verbose = True)


    #-get the NIfTI of registered image
    nim = nib.load(freg)
    A   = nim.affine
    imreg = np.float32( nim.get_data() )
    imreg = imreg[:,::-1,::-1]
    imreg = np.transpose(imreg, (2, 1, 0))

    #-convert to mu-values; sort out the file name too.
    if musrc=='pct':
        mu = hu2mu(imreg)
    elif musrc=='ute':
        mu = np.float32(imreg)/1e4
        #-remove the converted file from DICOMs
        os.remove(fflo)
    else:
        raise NameError('Confused o_O.')

    # get rid of negatives
    mu[mu<0] = 0
    # return image dictionary with the image itself and other parameters
    mu_dct['im'] = mu
    mu_dct['affine'] = A
    mu_dct['faff'] = faff
    
    if store or store_npy:

        nimpa.create_dir(opth)
        if faff=='':
            fname = fnm + '-aligned-to_t'\
                    + str(hst['t0'])+'-'+str(hst['t1'])+'_'+petopt.upper()\
                    + fcomment
        else:
            fname = fnm + '-aligned-to-given-affine' + fcomment

    if store_npy:
        #> Numpy
        if store_to_npy:
            fnp = os.path.join(opth, fname + '.npy')
            np.save(fnp, (mu, A, fnp))


    if store:
        #> NIfTI
        fmu = os.path.join(opth, fname + '.nii.gz')
        nimpa.array2nii(mu[::-1,::-1,:], A, fmu)
        mu_dct['fim'] = fmu

    if del_auxilary:
        os.remove(freg)
        if not os.path.isfile(faff):
            os.remove(fute)

    return mu_dct


#=================================================================================
# PSEUDO CT MU-MAP
#---------------------------------------------------------------------------------
def pct_mumap(
        datain, scanner_params,
        hst=[], t0=0, t1=0,
        itr=2,
        petopt='ac',
        faff='',
        fpet='',
        fcomment='',
        outpath='',
        store_npy = False,
        store = False, 
    ):
        
    '''
    GET THE MU-MAP from pCT IMAGE (which is in T1w space)
    * the mu-map will be registered to PET which will be reconstructed for time frame t0-t1
    * it f0 and t1 are not given the whole LM dataset will be reconstructed 
    * the reconstructed PET can be attenuation and scatter corrected or NOT using petopt
    '''

    if not os.path.isfile(faff):
        from niftypet.nipet.prj import mmrrec
        # histogram the list data if needed
        if not hst:
            from niftypet.nipet.lm import mmrhist
            hst = mmrhist.mmrhist(datain, scanner_params, t0=t0, t1=t1)
        
    # constants, transaxial and axial LUTs are extracted
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    # get hardware mu-map
    if 'hmumap' in datain and os.path.isfile(datain['hmumap']):
        muh, _, _ = np.load(datain['hmumap'])
        if Cnt['VERBOSE']: print 'i> loaded hardware mu-map from file:', datain['hmumap']
    elif outpath!='':
        hmupath = os.path.join( os.path.join(outpath,'mumap-hdw'), 'hmumap.npy')
        if os.path.isfile( hmupath ):
            muh, _, _ = np.load(hmupath)
            datain['hmumap'] = hmupath
        else:
            raise IOError('Invalid path to the hardware mu-map')
    else:
        print 'e> obtain the hardware mu-map first.'
        raise IOError('Could not find the hardware mu-map.  Have you run the routine for hardware mu-map?')

    if not 'MRT1W#' in datain and not 'T1nii' in datain and not 'T1bc' in datain:
        print 'e> no MR T1w images required for co-registration!'
        raise IOError('Missing MR data')
    # ----------------------------------

    # output dictionary
    mu_dct = {}
    if not os.path.isfile(faff):
        # first recon pet to get the T1 aligned to it
        if petopt=='qnt':
            # ---------------------------------------------
            # OPTION 1 (quantitative recon with all corrections using MR-based mu-map)
            # get UTE object mu-map (may not be in register with the PET data)
            mudic = obj_mumap(datain, Cnt)
            muo = mudic['im']
            # reconstruct PET image with UTE mu-map to which co-register T1w
            recout = mmrrec.osemone(   
                datain, [muh, muo], 
                hst, scanner_params,
                recmod=3, itr=itr, fwhm=0.,
                fcomment=fcomment+'_qntUTE',
                outpath=os.path.join(outpath, 'PET', 'positioning'),
                store_img=True)
        elif petopt=='nac':
            # ---------------------------------------------
            # OPTION 2 (recon without any corrections for scatter and attenuation)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            muo = np.zeros(muh.shape, dtype=muh.dtype)
            recout = mmrrec.osemone(
                datain, [muh, muo],
                hst, scanner_params, 
                recmod=1, itr=itr, fwhm=0., 
                fcomment=fcomment+'_NAC',
                outpath=os.path.join(outpath, 'PET', 'positioning'),
                store_img=True)
        elif petopt=='ac':
            # ---------------------------------------------
            # OPTION 3 (recon with attenuation correction only but no scatter)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            mudic = obj_mumap(datain, Cnt, outpath=outpath)
            muo = mudic['im']
            recout = mmrrec.osemone(
                datain, [muh, muo],
                hst, scanner_params, 
                recmod=1, itr=itr, fwhm=0., 
                fcomment=fcomment+'_AC',
                outpath=os.path.join(outpath, 'PET', 'positioning'),
                store_img=True)       

        fpet = recout.fpet
        mu_dct['fpet'] = fpet
        
        #------------------------------
        # get the affine transformation

        ft1w = nimpa.pick_t1w(datain)
        try:
            regdct = nimpa.coreg_spm(
                fpet,
                ft1w,
                outpath=os.path.join(outpath,'PET', 'positioning')
            )
        except:
            regdct = nimpa.affine_niftyreg(
                fpet,
                ft1w,
                outpath=os.path.join(outpath,'PET', 'positioning'),
                #fcomment=fcomment,
                executable = Cnt['REGPATH'],
                omp = multiprocessing.cpu_count()/2,
                rigOnly = True,
                affDirect = False,
                maxit=5,
                speed=True,
                pi=50, pv=50,
                smof=0, smor=0,
                rmsk=True,
                fmsk=True,
                rfwhm=15., #millilitres
                rthrsh=0.05,
                ffwhm = 15., #millilitres
                fthrsh=0.05,
                verbose=verbose
            )

        faff = regdct['faff']
        #------------------------------

    # pCT file name
    if outpath=='':
        pctdir = os.path.dirname(datain['pCT'])
    else:
        pctdir = os.path.join(outpath, 'mumap-obj')
    mmraux.create_dir(pctdir)
    fpct = os.path.join(pctdir, 'pCT_r_tmp'+fcomment+'.nii.gz')

    #call the resampling routine to get the pCT in place
    if os.path.isfile( Cnt['RESPATH'] ):
        cmd = [Cnt['RESPATH'],
            '-ref', fpet,
            '-flo', datain['pCT'],
            '-trans', faff,
            '-res', fpct,
            '-pad', '0']
        if not Cnt['VERBOSE']: cmd.append('-voff')
        call(cmd)
    else:
        print 'e> path to resampling executable is incorrect!'
        sys.exit()


    # get the NIfTI of the pCT
    nim = nib.load(fpct)
    A   = nim.get_sform()
    pct = np.float32( nim.get_data() )
    pct = pct[:,::-1,::-1]
    pct = np.transpose(pct, (2, 1, 0))
    # convert the HU units to mu-values
    mu = hu2mu(pct)
    # get rid of negatives
    mu[mu<0] = 0
    
    # return image dictionary with the image itself and other parameters
    mu_dct['im'] = mu
    mu_dct['affine'] = A
    mu_dct['faff'] = faff

    if store:
        # now save to numpy array and NIfTI in this folder
        if outpath=='':
            pctumapdir = os.path.join( datain['corepath'], 'mumap-obj' )
        else:
            pctumapdir = os.path.join(outpath, 'mumap-obj')
        mmraux.create_dir(pctumapdir)

        #> Numpy
        if store_npy:
            fnp = os.path.join(pctumapdir, 'mumap-pCT.npy')
            np.save(fnp, (mu, A, fnp))


        # numpy

        # NIfTI
        fmu = os.path.join(pctumapdir, 'mumap-pCT' +fcomment+ '.nii.gz')
        nimpa.array2nii(mu[::-1,::-1,:], A, fmu)
        mu_dct['fim'] = fmu
        datain['mumapCT'] = fmu

    return mu_dct


#*********************************************************************************
#GET HARDWARE MU-MAPS with positions and offsets
#---------------------------------------------------------------------------------
def hdr_mu(datain, Cnt):
    '''Get the headers from DICOM data file'''
    #get one of the DICOM files of the mu-map
    if 'mumapDCM' in datain:
        files = glob.glob(os.path.join(datain['mumapDCM'],'*.dcm'))
        files.extend(glob.glob(os.path.join(datain['mumapDCM'],'*.DCM')))
        files.extend(glob.glob(os.path.join(datain['mumapDCM'],'*.ima')))
        files.extend(glob.glob(os.path.join(datain['mumapDCM'],'*.IMA')))
        dcmf = files[0]
    else:
        print 'e> path to the DICOM mu-map is not given but it is required.'
        raise NameError('No DICOM mu-map')
    if os.path.isfile( dcmf ):
        dhdr = dcm.read_file( dcmf )
    else:
        print 'e> DICOM mMR mu-maps are not valid files!'
        return None
    # CSA Series Header Info
    if [0x29,0x1020] in dhdr:
        csahdr = dhdr[0x29,0x1020].value
        if Cnt['VERBOSE']: print 'i> got CSA mu-map info.'
    return csahdr, dhdr

def hmu_shape(hdr):
    #regular expression to find the shape
    p = re.compile(r'(?<=:=)\s*\d{1,4}')
    # x: dim [1]
    i0 = hdr.find('matrix size[1]')
    i1 = i0+hdr[i0:].find('\n')
    u = int(p.findall(hdr[i0:i1])[0])
    # x: dim [2]
    i0 = hdr.find('matrix size[2]')
    i1 = i0+hdr[i0:].find('\n')
    v = int(p.findall(hdr[i0:i1])[0])
    # x: dim [3]
    i0 = hdr.find('matrix size[3]')
    i1 = i0+hdr[i0:].find('\n')
    w = int(p.findall(hdr[i0:i1])[0])
    return w,v,u

def hmu_voxsize(hdr):
    #regular expression to find the shape
    p = re.compile(r'(?<=:=)\s*\d{1,2}[.]\d{1,10}')
    # x: dim [1]
    i0 = hdr.find('scale factor (mm/pixel) [1]')
    i1 = i0+hdr[i0:].find('\n')
    vx = float(p.findall(hdr[i0:i1])[0])
    # x: dim [2]
    i0 = hdr.find('scale factor (mm/pixel) [2]')
    i1 = i0+hdr[i0:].find('\n')
    vy = float(p.findall(hdr[i0:i1])[0])
    # x: dim [3]
    i0 = hdr.find('scale factor (mm/pixel) [3]')
    i1 = i0+hdr[i0:].find('\n')
    vz = float(p.findall(hdr[i0:i1])[0])
    return np.array([0.1*vz, 0.1*vy, 0.1*vx])

def hmu_origin(hdr):
    #regular expression to find the origin
    p = re.compile(r'(?<=:=)\s*\d{1,5}[.]\d{1,10}')
    # x: dim [1]
    i0 = hdr.find('$umap origin (pixels) [1]')
    i1 = i0+hdr[i0:].find('\n')
    x = float(p.findall(hdr[i0:i1])[0])
    # x: dim [2]
    i0 = hdr.find('$umap origin (pixels) [2]')
    i1 = i0+hdr[i0:].find('\n')
    y = float(p.findall(hdr[i0:i1])[0])
    # x: dim [3]
    i0 = hdr.find('$umap origin (pixels) [3]')
    i1 = i0+hdr[i0:].find('\n')
    z = -float(p.findall(hdr[i0:i1])[0])
    return np.array([z, y, x])

def hmu_offset(hdr):
    #regular expression to find the origin
    p = re.compile(r'(?<=:=)\s*\d{1,5}[.]\d{1,10}')
    if hdr.find('$origin offset')>0:
        # x: dim [1]
        i0 = hdr.find('$origin offset (mm) [1]')
        i1 = i0+hdr[i0:].find('\n')
        x = float(p.findall(hdr[i0:i1])[0])
        # x: dim [2]
        i0 = hdr.find('$origin offset (mm) [2]')
        i1 = i0+hdr[i0:].find('\n')
        y = float(p.findall(hdr[i0:i1])[0])
        # x: dim [3]
        i0 = hdr.find('$origin offset (mm) [3]')
        i1 = i0+hdr[i0:].find('\n')
        z = -float(p.findall(hdr[i0:i1])[0])
        return np.array([0.1*z, 0.1*y, 0.1*x])
    else:
        return np.array([0.0, 0.0, 0.0])

def rd_hmu(fh):
    #--read hdr file--
    f = open(fh, 'r')
    hdr = f.read()
    f.close()
    #-----------------
    #regular expression to find the file name
    p = re.compile(r'(?<=:=)\s*\w*[.]\w*')
    i0 = hdr.find('!name of data file')
    i1 = i0+hdr[i0:].find('\n')
    fbin = p.findall(hdr[i0:i1])[0]
    #--read img file--
    f = open(os.path.join(os.path.dirname(fh), fbin.strip()), 'rb')
    im = np.fromfile(f, np.float32)
    f.close()
    #-----------------
    return  hdr, im 


def get_hmupos(datain, parts, Cnt, outpath=''):

    # check if registration executable exists
    if not os.path.isfile(Cnt['RESPATH']):
        print 'e> no registration executable found!'
        sys.exit()

    #----- get positions from the DICOM list-mode file -----
    ihdr, csainfo = mmraux.hdr_lm(datain, Cnt)
    #table position origin
    fi = csainfo.find('TablePositionOrigin')
    tpostr = csainfo[fi:fi+200]
    tpo = re.sub(r'[^a-zA-Z0-9\-\.]', '', tpostr).split('M')
    tpozyx = np.array([float(tpo[-1]), float(tpo[-2]), float(tpo[-3])]) / 10
    if Cnt['VERBOSE']: print 'i> table position (z,y,x) (cm):', tpozyx
    #--------------------------------------------------------

    #------- get positions from the DICOM mu-map file -------
    csamu, dhdr = hdr_mu(datain, Cnt)
    tmp = re.search('GantryTableHomeOffset(?!_)', csamu)
    gtostr1  = csamu[ tmp.start():tmp.start()+300 ]
    gtostr2 = re.sub(r'[^a-zA-Z0-9\-\.]', '', gtostr1)
    # gantry table offset, through conversion of string to float
    gtoxyz = re.findall(r'(?<=M)-*[\d]{1,4}\.[\d]{6,9}', gtostr2)
    gtozyx = np.float32(gtoxyz)[::-1]/10
    #--------------------------------------------------------

    if Cnt['VERBOSE']: print 'i> gantry table offset (z,y,x) (cm):', gtozyx

    ## ----
    ## old II
    # csamu, dhdr = nipet.img.mmrimg.hdr_mu(datain, Cnt)
    # tmp = re.search('GantryTableHomeOffset(?!_)', csamu)
    # gtostr = csamu[ tmp.start():tmp.start()+300 ]
    # gto = re.sub(r'[^a-zA-Z0-9\-\.]', '', gtostr).split('M')
    # # get the first three numbers
    # zyx = np.zeros(3, dtype=np.float32)
    # c = 0
    # for i in range(len(gto)):
    #     if re.search(r'[\d]{1,3}\.[\d]{6}', gto[i])!=None and c<3:
    #         zyx[c] = np.float32(re.sub(r'[^0-9\-\.]', '', gto[i]))
    #         c+=1
    # #gantry table offset
    # gtozyx = zyx[::-1]/10
    ## ----

    ## ----
    ## old I: only worked for syngo MR B20P
    # fi = csamu.find('GantryTableHomeOffset') 
    # gtostr =csamu[fi:fi+300]
    # if dhdr[0x0018, 0x1020].value == 'syngo MR B20P':
    #     gto = re.sub(r'[^a-zA-Z0-9\-\.]', '', gtostr).split('M')
    #     # get the first three numbers
    #     zyx = np.zeros(3, dtype=np.float32)
    #     c = 0
    #     for i in range(len(gto)):
    #         if re.search(r'[\d]', gto[i])!=None and c<3:
    #             zyx[c] = np.float32(re.sub(r'[^0-9\-\.]', '', gto[i]))
    #             c+=1
    #     #gantry table offset
    #     gtozyx = zyx[::-1]/10
    #     if Cnt['VERBOSE']: print 'i> gantry table offset (z,y,x) (cm):', gtozyx
    # # older scanner version
    # elif dhdr[0x0018, 0x1020].value == 'syngo MR B18P':
    #     zyx = np.zeros(3, dtype=np.float32)
    #     for k in range(3):
    #         tmp = re.search(r'\{\s*[\-0-9.]*\s*\}', gtostr)
    #         i0 = tmp.start()
    #         i1 = tmp.end()
    #         if gtostr[i0+1:i1-1]!=' ':  zyx[k] = np.float32(gtostr[i0+1:i1-1])
    #         gtostr = gtostr[i1:]
    #     #gantry table offset
    #     gtozyx = zyx[::-1]/10
    #     if Cnt['VERBOSE']: print 'i> gantry table offset (z,y,x) (cm):', gtozyx
    ## -----

    # create the folder for hardware mu-maps
    if outpath=='':
        dirhmu = os.path.join( datain['corepath'], 'mumap-hdw')
    else:
        dirhmu = os.path.join( outpath, 'mumap-hdw')
    mmraux.create_dir(dirhmu)
    # get the reference nii image
    fref = os.path.join(dirhmu, 'hmuref.nii.gz')

    #start horizontal bed position
    p = re.compile(r'start horizontal bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    hbedpos = 0.1*float(ihdr[m.start()+fi+1:m.end()])

    #start vertical bed position
    p = re.compile(r'start vertical bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    vbedpos = 0.1*float(ihdr[m.start()+fi+1:m.end()])

    if Cnt['VERBOSE']: print 'i> creating reference nii image for resampling'
    B = np.diag(np.array([-10*Cnt['SO_VXX'], 10*Cnt['SO_VXY'], 10*Cnt['SO_VXZ'], 1]))
    B[0,3] = 10*(.5*Cnt['SO_IMX'])*Cnt['SO_VXX']
    B[1,3] = 10*( -.5*Cnt['SO_IMY']+1)*Cnt['SO_VXY']
    B[2,3] = 10*((-.5*Cnt['SO_IMZ']+1)*Cnt['SO_VXZ'] + hbedpos )
    nimpa.array2nii(  np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32), B, fref)

    #define a dictionary of all positions/offsets of hardware mu-maps
    hmupos = [None]*5
    hmupos[0] = {
        'TabPosOrg' :   tpozyx, #from DICOM of LM file
        'GanTabOff' :   gtozyx, #from DICOM of mMR mu-map file
        'HBedPos'   :   hbedpos, #from Interfile of LM file [cm]
        'VBedPos'   :   vbedpos, #from Interfile of LM file [cm]
        'niipath'   :   fref
        }

    #--------------------------------------------------------------------------
    # iteratively go through the mu-maps and add them as needed
    for i in parts:
        fh = os.path.join(Cnt['HMUDIR'], Cnt['HMULIST'][i-1])
        # get the interfile header and binary data 
        hdr, im = rd_hmu(fh)
        #get shape, origin, offset and voxel size
        s = hmu_shape(hdr)
        im.shape = s
        # get the origin, offset and voxel size for the mu-map interfile data
        org = hmu_origin(hdr)
        off = hmu_offset(hdr)
        vs  = hmu_voxsize(hdr)
        # corner voxel position for the interfile image data
        vpos = (-org*vs + off + gtozyx - tpozyx)
        #add to the dictionary
        hmupos[i] = {
            'vpos'    :   vpos,
            'shape'   :   s,   #from interfile
            'iorg'    :   org, #from interfile
            'ioff'    :   off, #from interfile
            'ivs'     :   vs,  #from interfile
            'img'     :   im, #from interfile
            'niipath' :   os.path.join(dirhmu, '_'+Cnt['HMULIST'][i-1].split('.')[0]+'.nii.gz')
        }
        #save to NIfTI
        if Cnt['VERBOSE']: print 'i> creating mu-map for:', Cnt['HMULIST'][i-1]
        A = np.diag(np.append(10*vs[::-1], 1))
        A[0,0] *= -1
        A[0,3] =  10*(-vpos[2])
        A[1,3] = -10*((s[1]-1)*vs[1] + vpos[1])
        A[2,3] = -10*((s[0]-1)*vs[0] - vpos[0])
        nimpa.array2nii(im[::-1,::-1,:], A, hmupos[i]['niipath'])

        # resample using nify.reg
        fout = os.path.join(    os.path.dirname (hmupos[0]['niipath']),
                                'r'+os.path.basename(hmupos[i]['niipath']).split('.')[0]+'.nii.gz' )
        cmd = [ Cnt['RESPATH'],
                '-ref', hmupos[0]['niipath'],
                '-flo', hmupos[i]['niipath'],
                '-res', fout,
                '-pad', '0']
        if not Cnt['VERBOSE']: cmd.append('-voff')
        call(cmd)
                
    return hmupos


#-------------------------------------------------------------------------------------
def hdw_mumap(
        datain, 
        hparts, 
        params, 
        outpath='', 
        use_stored=False,
        del_interm=True):
    ''' Get hardware mu-map components, including bed, coils etc.
    '''

    # two ways of passing Cnt are here decoded
    if 'Cnt' in params:
        Cnt = params['Cnt']
    else:
        Cnt = params

    if outpath!='':
        fmudir = os.path.join(outpath, 'mumap-hdw')
    else:
        fmudir = os.path.join(datain['corepath'], 'mumap-hdw')

    # if requested to use the stored hardware mu_map get it from the path in datain
    if 'hmumap' in datain and os.path.isfile(datain['hmumap']) and use_stored:

        if datain['hmumap'].endswith(('.nii', '.nii.gz')):
            dct = nimpa.getnii(datain['hmumap'], output='all')
            hmu = dct['im']
            A = dct['affine']
            fmu = datain['hmumap']
        elif datain['hmumap'].endswith(('.npy')):
            hmu, A, fmu = np.load(datain['hmumap'])
            if Cnt['VERBOSE']: print 'i> loaded hardware mu-map from file:', datain['hmumap']
            fnp = datain['hmumap']

    elif outpath!='' and os.path.isfile(os.path.join(fmudir, 'hmumap.npy')):
        fnp = os.path.join(fmudir, 'hmumap.npy')
        hmu, A, fmu = np.load(fnp)
        datain['hmumap'] = fnp
    # otherwise generate it from the parts through resampling the high resolution CT images
    else:
        hmupos = get_hmupos(datain, hparts, Cnt, outpath=outpath)
        # just to get the dims, get the ref image
        nimo = nib.load(hmupos[0]['niipath'])
        A = nimo.get_sform()
        imo = np.float32( nimo.get_data() )
        imo[:] = 0

        for i in hparts:
            fin  = os.path.join(os.path.dirname (hmupos[0]['niipath']),
                                'r'+os.path.basename(hmupos[i]['niipath']).split('.')[0]+'.nii.gz' )
            nim = nib.load(fin)
            mu = nim.get_data()
            mu[mu<0] = 0
            
            imo += mu

        hdr = nimo.header
        hdr['cal_max'] = np.max(imo)
        hdr['cal_min'] = np.min(imo)
        fmu  = os.path.join(os.path.dirname (hmupos[0]['niipath']), 'hardware_umap.nii.gz' )
        nib.save(nimo, fmu)

        hmu = np.transpose(imo[:,::-1,::-1], (2, 1, 0))

        # save the objects to numpy arrays
        fnp = os.path.join(fmudir, 'hmumap.npy')
        np.save(fnp, (hmu, A, fmu))
        #update the datain dictionary (assuming it is mutable)
        datain['hmumap'] = fnp

        if del_interm:
            for fname in glob.glob(os.path.join(fmudir, '_*.nii*')):
                os.remove(fname)
            for fname in glob.glob(os.path.join(fmudir, 'r_*.nii*')):
                os.remove(fname)

    #return image dictionary with the image itself and some other stats
    hmu_dct = { 'im':hmu,
                'fim':fmu,                
                'affine':A}
    if 'fnp' in locals():
        hmu_dct['fnp'] = fnp

    return hmu_dct


def rmumaps(datain, Cnt, t0=0, t1=0, use_stored=False):
    '''
    get the mu-maps for hardware and object and trim it axially for reduced rings case
    '''

    from niftypet.nipet.lm  import mmrhist
    from niftypet.nipet.prj import mmrrec

    fcomment = '(R)'

    # get hardware mu-map
    if os.path.isfile(datain['hmumap']) and use_stored:
        muh, _ = np.load(datain['hmumap'])
        if Cnt['VERBOSE']: print 'i> loaded hardware mu-map from file:', datain['hmumap']
    else:
        hmudic = hdw_mumap(datain, [1,2,4], Cnt)
        muh = hmudic['im']

    # get pCT mu-map if stored in numpy file and then exit, otherwise do all the processing
    if os.path.isfile(datain['mumapCT']) and use_stored:
        mup, _ = np.load(datain['mumapCT'])
        muh = muh[2*Cnt['RNG_STRT'] : 2*Cnt['RNG_END'], :, :]
        mup = mup[2*Cnt['RNG_STRT'] : 2*Cnt['RNG_END'], :, :]
        return [muh, mup]

    # get UTE object mu-map (may be not in register with the PET data)
    if os.path.isfile(datain['mumapUTE']) and use_stored:
        muo, _ = np.load(datain['mumapUTE'])
    else:
        mudic = obj_mumap(datain, Cnt, store=True)
        muo = mudic['im']

    if os.path.isfile(datain['pCT']):
        # reconstruct PET image with default settings to be used to alight pCT mu-map
        params = mmraux.mMR_params()
        Cnt_ = params['Cnt']
        txLUT_ = params['txLUT']
        axLUT_ = params['axLUT']
        
        # histogram for reconstruction with UTE mu-map
        hst = mmrhist.hist(datain, txLUT_, axLUT_, Cnt_, t0=t0, t1=t1)
        # reconstruct PET image with UTE mu-map to which co-register T1w
        recute = mmrrec.osemone(
            datain, [muh, muo], hst, params, 
            recmod=3, itr=4, fwhm=0., store_img=True, fcomment=fcomment+'_QNT-UTE'
        )
        # --- MR T1w
        if os.path.isfile(datain['T1nii']):
            ft1w = datain['T1nii']
        elif os.path.isfile(datain['T1bc']):
            ft1w = datain['T1bc']
        elif os.path.isdir(datain['MRT1W']):
            # create file name for the converted NIfTI image
            fnii = 'converted'
            call( [ Cnt['DCM2NIIX'], '-f', fnii, datain['T1nii'] ] )
            ft1nii = glob.glob( os.path.join(datain['T1nii'], '*converted*.nii*') )
            ft1w = ft1nii[0]
        else:
            print 'e> disaster: no T1w image!'
            sys.exit()

        #output for the T1w in register with PET
        ft1out = os.path.join(os.path.dirname(ft1w), 'T1w_r'+'.nii.gz')
        #text file fo rthe affine transform T1w->PET
        faff   = os.path.join(os.path.dirname(ft1w), fcomment+'mr2pet_affine'+'.txt')  #time.strftime('%d%b%y_%H.%M',time.gmtime())
        #call the registration routine
        if os.path.isfile( Cnt['REGPATH'] ):
            cmd = [Cnt['REGPATH'],
                 '-ref', recute.fpet,
                 '-flo', ft1w,
                 '-rigOnly', '-speeeeed',
                 '-aff', faff,
                 '-res', ft1out]
            if not Cnt['VERBOSE']: cmd.append('-voff')
            call(cmd)
        else:
            print 'e> path to registration executable is incorrect!'
            sys.exit()

        #get the pCT mu-map with the above faff
        pmudic = pct_mumap(datain, txLUT, axLUT, Cnt, faff=faff, fpet=recute.fpet, fcomment=fcomment)
        mup = pmudic['im']

        muh = muh[2*Cnt['RNG_STRT'] : 2*Cnt['RNG_END'], :, :]
        mup = mup[2*Cnt['RNG_STRT'] : 2*Cnt['RNG_END'], :, :]
        return [muh, mup]
    else:
        muh = muh[2*Cnt['RNG_STRT'] : 2*Cnt['RNG_END'], :, :]
        muo = muo[2*Cnt['RNG_STRT'] : 2*Cnt['RNG_END'], :, :]
        return [muh, muo]

#------------------------------------------------------------------------



"""mmraux.py: auxilary functions for PET list-mode data processing."""

__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"

#-------------------------------------------------------------------------------
import numpy as np
import sys
import os
import pydicom as dcm
import re
from pkg_resources import resource_filename


#auxiliary functions through Python extensions in CUDA
import mmr_auxe


#=================================================================================================
# GET NORM COMPONENTS
#=================================================================================================
def get_components(datain, Cnt):
    "Return the normalisation components from provided file."
    
    if 'nrm_ima' in datain and os.path.isfile(datain['nrm_ima']):
        fnrm_dat = datain['nrm_bf']
        fnrm_hdr = datain['nrm_ima']

    elif 'nrm_dcm' in datain and os.path.isfile(datain['nrm_dcm']) and os.path.isfile(datain['nrm_bf']):
        fnrm_dat = datain['nrm_bf']
        fnrm_hdr = datain['nrm_dcm']
    else:
        print 'e> norm file does not exist or it is incomplete.'
        raise NameError('NoNorm')

    f = open(fnrm_dat, 'rb')
    #geometric effects
    geo       = np.fromfile(f, np.float32, Cnt['NSBINS']*Cnt['NSEG0'])
    geo.shape = (Cnt['NSEG0'], Cnt['NSBINS'])
    #crystal interference
    crs_intf  = np.fromfile(f, np.float32, 9*Cnt['NSBINS'])
    crs_intf.shape = (Cnt['NSBINS'],9)
    #crystal efficiencies
    crs_eff   = np.fromfile(f, np.float32, Cnt['NCRS']*Cnt['NRNG'])
    crs_eff.shape  = (Cnt['NRNG'], Cnt['NCRS'])
    #axial effects
    ax_eff1   = np.fromfile(f, np.float32, Cnt['NSN11'])
    #paralyzing ring DT parameters
    rng_dtp   = np.fromfile(f, np.float32, Cnt['NRNG'])
    #non-paralyzing ring DT parameters
    rng_dtnp  = np.fromfile(f, np.float32, Cnt['NRNG'])
    #TX crystal DT parameter
    crs_dt    = np.fromfile(f, np.float32, 9)
    #additional axial effects
    ax_eff2   = np.fromfile(f, np.float32, Cnt['NSN11'])
    f.close()

    #-------------------------------------------------
    #the files below are found based on a 24hr scan of germanium-68 phantom
    # axial effects for span-1
    fnm = 'AxialFactorForSpan1.npy'
    fpth = os.path.join(resource_filename('niftypet', 'auxdata'), fnm)
    ax_f1 = np.load(fpth)
    # relative scale factors for axial scatter deriving span-11 scale factors from SSR scale factors
    fnm = 'RelativeScaleFactors_scatter_axial_ssrTOspan11.f32'
    fpth = os.path.join(resource_filename('niftypet', 'auxdata'), fnm)
    f = open(fpth, 'rb')
    sax_f11 = np.fromfile(f, np.float32, Cnt['NSN11'])
    f.close()
    # relative scale factors for axial scatter deriving span-1 scale factors from SSR scale factors
    fnm = 'RelativeScaleFactors_scatter_axial_ssrTOspan1.f32'
    fpth = os.path.join(resource_filename('niftypet', 'auxdata'), fnm)
    f = open(fpth, 'rb')
    sax_f1 = np.fromfile(f, np.float32, Cnt['NSN1'])
    f.close()
    #-------------------------------------------------

    #-------------------------------------------------
    # HEADER FILE
    # possible DICOM locations for the Interfile header
    nhdr_locations = [[0x29,0x1010], [0x29,0x1110]]
    # read the DICOM file
    d = dcm.read_file(fnrm_hdr)
    if   d[0x0018, 0x1020].value == 'syngo MR B20P' or d[0x0018, 0x1020].value == 'syngo MR E11':
        nhdr = d[0x29,0x1010].value
    elif d[0x0018, 0x1020].value == 'syngo MR B18P':
        found_nhdr = False
        for loc in nhdr_locations:
            if loc in d:
                nhdr = d[loc].value
                if '!INTERFILE' in nhdr and 'scanner quantification factor' in nhdr:
                    if Cnt['VERBOSE']: print 'i> got the normalisation interfile header from [', hex(loc[0]),',', hex(loc[1]), ']'
                    found_nhdr = True
                    break
        if not found_nhdr:           
            print 'e> DICOM field with normalisation interfile header has not been found!'
            return None, None
    else: print 'e> unknown scanner software version!';  return None, None

    f0 = nhdr.find('scanner quantification factor')
    f1 = f0+nhdr[f0:].find('\n')
    #regular expression for the needed three numbers
    p = re.compile(r'(?<=:=)\s*\d{1,5}[.]\d{3,10}[e][+-]\d{1,4}')
    #-quantification factor:
    qf = float(p.findall(nhdr[f0:f1])[0])
    #-local quantification correction factor
    qf_loc = 0.205
    #-------------------------------------------------

    nrmcmp = {'qf':qf, 'qf_loc':qf_loc, 'geo':geo, 'cinf':crs_intf, 'ceff':crs_eff,
                'axe1':ax_eff1, 'dtp':rng_dtp, 'dtnp':rng_dtnp,
                'dtc':crs_dt, 'axe2':ax_eff2, 'axf1':ax_f1, 
                'sax_f11':sax_f11, 'sax_f1':sax_f1}
    

    return nrmcmp, nhdr


def get_sinog(datain, hst, axLUT, txLUT, Cnt, normcomp=None):

    #get the normalisation components
    if normcomp==None:
        normcomp, _ = get_components(datain, Cnt)

    #number of sino planes (2D sinos) depends on the span used
    if Cnt['SPN']==1:
        nsinos = Cnt['NSN1']
    elif Cnt['SPN']==11:
        nsinos = Cnt['NSN11']

    #predefine the sinogram
    sinog = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)

    #get the sino in the GPU-optimised shape
    mmr_auxe.norm(sinog, normcomp, hst['buckets'], axLUT, txLUT['aw2ali'], Cnt)

    return sinog


def get_sino(datain, hst, axLUT, txLUT, Cnt):

    #number of sino planes (2D sinos) depends on the span used
    if Cnt['SPN']==1:
        nsinos = Cnt['NSN1']
    elif Cnt['SPN']==11:
        nsinos = Cnt['NSN11']

    #get sino with no gaps
    s = get_sinog(datain, hst, axLUT, txLUT, Cnt)
    #preallocate sino with gaps
    sino = np.zeros((Cnt['NSANGLES'], Cnt['NSBINS'], nsinos), dtype=np.float32)
    #fill the sino with gaps
    mmr_auxe.pgaps(sino, s, txLUT, Cnt)
    sino = np.transpose(sino, (2,0,1))

    return sino

def get_norm_sino(datain, scanner_params, hst):

    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    # if not hst:
    #     hst = mmrhist.mmrhist(datain, scanner_params)
        
    #number of sino planes (2D sinos) depends on the span used
    if Cnt['SPN']==1:
        nsinos = Cnt['NSN1']
    elif Cnt['SPN']==11:
        nsinos = Cnt['NSN11']

    #get sino with no gaps
    s = get_sinog(datain, hst, axLUT, txLUT, Cnt)
    #preallocate sino with gaps
    sino = np.zeros((Cnt['NSANGLES'], Cnt['NSBINS'], nsinos), dtype=np.float32)
    #fill the sino with gaps
    mmr_auxe.pgaps(sino, s, txLUT, Cnt)
    sino = np.transpose(sino, (2,0,1))

    return sino



#!/usr/bin/python
__author__ = 'pawel'

#--- XNAT ---
import sys
import os
import csv
import shutil
#--------------

#--------------
import pycurl
import json
from StringIO import StringIO
from urllib import urlencode
import datetime
#--------------

#--------------
# import matplotlib
# # switch off X-windows (used for storing in png format)
# matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy as np
import math
import scipy.ndimage as ndi
#--------------


#-------------- INIT mMR ----------------
import nipet
import pviews
#-----------------------------------------


def time_stamp():
    now    = datetime.datetime.now()
    nowstr = str(now.year)+'-'+str(now.month)+'-'+str(now.day)+' '+str(now.hour)+':'+str(now.minute)
    return nowstr

def create_dir(pth):
    if not os.path.exists(pth):
        os.makedirs(pth)


#----------------------------------------------------------------------------------------------------------
def get_xnatList(usrpwd, xnaturi, cookie=''):
    buffer = StringIO()
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    else:
        c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.WRITEDATA, buffer)
    c.perform()
    c.close()
    # convert to json dictionary in python
    outjson = json.loads( buffer.getvalue() )
    return outjson['ResultSet']['Result']

def get_xnat(usrpwd, xnaturi, cookie='', frmt='json'):
    buffer = StringIO()
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    else:
        c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.WRITEDATA, buffer)
    c.perform()
    c.close()
    # convert to json dictionary in python
    if frmt=='':
        output = buffer.getvalue()
    elif frmt=='json':
        output = json.loads( buffer.getvalue() )
    return output

def get_xnatFile(usrpwd, xnaturi, fname, cookie=''):
    try:
        fn = open(fname, 'wb')
        c = pycurl.Curl()
        if cookie:
            c.setopt(pycurl.COOKIE, cookie)
        else:
            c.setopt(c.USERPWD, usrpwd)
        c.setopt(pycurl.SSL_VERIFYPEER, 0)
        c.setopt(pycurl.SSL_VERIFYHOST, 0)
        c.setopt(c.VERBOSE, 0)
        c.setopt(c.URL, xnaturi )
        c.setopt(c.WRITEDATA, fn)
        c.setopt(pycurl.FOLLOWLOCATION, 0)
        c.setopt(pycurl.NOPROGRESS, 0)
        c.perform()
        c.close()
        fn.close()
    except pycurl.error as pe:
        print ' '
        print '=============================================================='
        print 'e> pycurl error:', pe
        print '=============================================================='
        print ' '
        print 'w> no data:', sbjid, sbjlb
        return -1
    else:
        print '--------'
        print 'i> pycurl download done.'
        print '--------'
    return 0
#----------------------------------------------------------------------------------------------------------


#----------------------------------------------------------------------------------------------------------
def put_xnat(usrpwd, xnaturi, cookie=''):
    """e.g., create a container"""
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    else:
        c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.CUSTOMREQUEST, 'PUT')
    c.perform()
    c.close()

def del_xnat(usrpwd, xnaturi, cookie=''):
    """e.g., create a container"""
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    else:
        c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.CUSTOMREQUEST, 'DELETE')
    c.perform()
    c.close()

def post_xnat(usrpwd, xnaturi, post_data, verbose=0, PUT=False):
    buffer = StringIO()
    c = pycurl.Curl()
    c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, verbose)
    c.setopt(c.URL, xnaturi )
    if PUT: c.setopt(c.CUSTOMREQUEST, 'PUT')
    c.setopt(c.POSTFIELDS, post_data)
    c.setopt(c.WRITEFUNCTION, buffer.write)
    c.perform()
    c.close()
    return buffer.getvalue()

def put_xnatFile(usrpwd, xnaturi, filepath, cookie=''):
    """upload file to xnat server"""
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    else:
        c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(pycurl.NOPROGRESS, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.HTTPPOST, [('fileupload', (c.FORM_FILE, filepath,)),])
    c.perform()
    c.close()
#----------------------------------------------------------------------------------------------------------



#----------------------------------------------------------------------------------------------------------
def put_xnatPetMrRes(usrpwd, xnatsbj, sbjix, lbl, frmt, fpth):
    # get the experiment id
    eid = get_xnatList(usrpwd, xnatsbj+'/' +sbjix+ '/experiments?xsiType=xnat:petmrSessionData&format=json&columns=ID')
    # prepare the uri
    xnaturi = xnatsbj+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/resources/' +lbl+ '?xsi:type=xnat:resourceCatalog&format='+frmt
    xnat_put(usrpwd, xnaturi)
    # upload
    xnaturi = xnatsbj+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/resources/' +lbl+ '/files'
    xnat_upload(usrpwd, xnaturi, fpth)



#----------------------------------------------------------------------------------------------------------
def pltQC(hst, outpth, Cnt):
    T = hst['dur']
    # plot headcurve:
    plt.figure(0)
    prm, = plt.plot(range(T), hst['phc']/1000, 'k', label='prompts')
    rnd, = plt.plot(range(T), hst['dhc']/1000, 'r', label='delayeds')
    plt.grid(b=True, which='major', color='k', linestyle=':', linewidth=.5)
    plt.ylabel('kcounts')
    plt.xlabel('seconds')
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=2, mode="expand", borderaxespad=0.)
    plt.savefig(os.path.join(outpth, 'headcurve.png'), format='png', dpi=300)

    # plot centre of mass:
    plt.figure(1)
    scmass = ndi.filters.gaussian_filter(hst['cmass'], 5, mode='mirror')
    plt.plot(scmass,'k')
    plt.grid(b=True, which='major', color='k', linestyle=':', linewidth=.5)
    plt.ylabel('cm')
    plt.xlabel('seconds')
    plt.savefig(os.path.join(outpth, 'mass_centre.png'), format='png', dpi=300)
    # ---
    #find the peak in hc to get the zooming-in range
    # i = np.argmax(hst['phc'])
    # ymin = np.floor( min(hst['cmass'][i:i+300]) )
    # ymax = np.ceil( max(hst['cmass'][i+100:]) )

    # prompt peak
    HCpeakIdx = np.nanargmax(hst['phc'])
    # the centre of mass peak has to be just around the prompt peak
    CMpeakIdx = np.nanargmax(hst['cmass'][:HCpeakIdx+60])
    # lowest point after the peak
    lcmss = np.nanmin(hst['cmass'][CMpeakIdx:])
    # get the index of the point
    il = CMpeakIdx + np.nanargmin(hst['cmass'][CMpeakIdx:])
    ymin = np.floor(lcmss*10)/10
    ymax = np.nanmax(hst['cmass'][il:])
    # range
    Rng = ymax-ymin
    if Rng<0.5:  ymax += 0.5-Rng
    plt.ylim([ymin, ymax])
    plt.xlim([0,hst['dur']])
    plt.savefig(os.path.join(outpth, 'mass_centre_zoom.png'), format='png', dpi=300)


#----------------------------------------------------------------------------------------------------------




#==========================================================================================================
# DOWNLOAD XNAT DATA AND PROCESS
#----------------------------------------------------------------------------------------------------------
def processData(xc, ip, sbj, plist, frms, txLUT, axLUT, Cnt, xnat_upload=False):
    """download then process the LM/norm data and store"""

    # ip: index of previously processed subject list (if exists)
    # sbj: subject attributes
    # opth: output path
    # plist: list of processed subjects (uses ip index)

    sbjix = sbj['ID']
    sbjid = sbj['ID'][9:]
    sbjlb = sbj['label']
    idate = sbj['insert_date']

    #core path
    core_path = os.path.join(xc['opth'], sbjid+'_'+sbjlb)
    # norm path
    npth = os.path.join(core_path, 'norm')
    # LM path
    lmpth = os.path.join(core_path, 'LM')
    # processing output path
    outpth = os.path.join(core_path, 'output')
    # destination output files
    qcfiles = os.path.join(xc['qcpth'], sbjid +'_'+ sbjlb)

    # output file names and their xnat content tags
    qccharts = [['headcurve.png','mass_centre.png','mass_centre_zoom.png'],['CURVE','COM','COMZ']]
    qcvideos = [['pView_dyn.mp4','pViews_6.mp4'],['PVD','PVS']]



    # get the experiment ID
    try:
        eid = get_xnatList(xc['usrpwd'], xc['sbjs']+'/' +sbjix+ '/experiments?xsiType=xnat:petmrSessionData&format=json&columns=ID')
    except ValueError as ve:
        print '-------------------'
        print 'e> Value Error:', ve
        print '   for subject:', sbjix
        print '-------------------'
    # if no experiments, just skip the rest
    if not eid:
        print 'w> no data for:', sbjid, sbjlb
        if ip>=0: del plist[ip]
        plist.append([sbjid, sbjlb, idate, 'n', time_stamp(), 'n', time_stamp()])

        # return empty stuff
        return []

    else:
        create_dir(npth)
        create_dir(lmpth)
        create_dir(outpth)

        # get the list of norm files
        nrmfiles = get_xnatList(xc['usrpwd'], xc['sbjs']+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/resources/Norm/files')
        # download the norm files:
        for i in range(len(nrmfiles)):
            status = get_xnatFile( xc['usrpwd'], xc['url']+nrmfiles[i]['URI'], os.path.join(npth, nrmfiles[i]['Name']) )
            if status<0:
                if ip>=0: del plist[ip]
                print 'w> no NORM data:', sbjid, sbjlb
                plist.append([sbjid, sbjlb, idate, 'n', time_stamp(), 'n', time_stamp()])
                return []

        # get the list of LM files
        lmfiles = get_xnatList(xc['usrpwd'], xc['sbjs']+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/resources/LM/files')
        # download the LM files:
        for i in range(len(lmfiles)):
            # check if the file is already downloaded:
            if  os.path.isfile ( os.path.join(lmpth, lmfiles[i]['Name']) ) and \
                str( os.path.getsize( os.path.join(lmpth, lmfiles[i]['Name']) ) )==lmfiles[i]['Size']:
                    print 'i> file of the same size,',lmfiles[i]['Name'], 'already exists: skipping download.'
            else:
                status = get_xnatFile( xc['usrpwd'], xc['url']+lmfiles[i]['URI'], os.path.join(lmpth, lmfiles[i]['Name']) )
                if status<0:
                    if ip>=0: del plist[ip]
                    print 'w> no LM data:', sbjid, sbjlb
                    plist.append([sbjid, sbjlb, idate, 'n', time_stamp(), 'n', time_stamp()])
                    return []

        #------------------------------------------------------
        if ip>=0: del plist[ip]

        # explore the input
        folderin = os.path.dirname(lmpth)
        datain = nipet.mmraux.explore_input(folderin, Cnt)
        # check if all data is present
        data_required = [['lm_dcm', 'lm_bf', 'nrm_dcm', 'nrm_bf'],
                        ['lm_ima', 'nrm_ima']]
        dcm_chck = [i in datain for i in data_required[0]]
        ima_chck = [i in datain for i in data_required[1]]
        if all(dcm_chck) or all(ima_chck):
            print 'i> all data present.'
        else:
            print 'e> data missing for', sbjid, sbjlb
            plist.append([sbjid, sbjlb, idate, 'n', time_stamp(), 'n', time_stamp()])
            return []
        #------------------------------------------------------
        

        # process the LM data:
        for i in range(len(lmfiles)):
            if os.path.splitext(lmfiles[i]['Name'])[1]=='.bf':
                # =================================================================================processing---------
                #-get basic time info of the LM data:
                nele, ttags, addrs = nipet.lm.mmr_lmproc.lminfo(os.path.join(lmpth, lmfiles[i]['Name']))
                nitag = (ttags[1]-ttags[0]+999)/1000
                print 'i> duration by integrating time tags [s]:', nitag

                #time difference between acquisition and norm QC
                nipet.mmraux.time_diff_norm_acq(datain)
                #histogram the LM data
                hst = nipet.lm.mmrhist.hist(datain, txLUT, axLUT, Cnt, frms=frms)
                #get video views of the projection data
                pviews.video_dyn11(hst, frms, outpth, axLUT, Cnt)
                pviews.video_frm(hst, outpth, Cnt)
                #-plot the head-curve and centre of mass
                plt.close('all')
                pltQC(hst, outpth, Cnt)

                # XNAT UPLOAD
                if xnat_upload:
                    # store the QC attributes to xnat
                    xnatAsses = xc['sbjs']+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/assessors/LMQC_' +sbjid+ '_' +sbjlb

                    try:
                        put_xnat(xc['usrpwd'], xnatAsses+'?xsiType=avid:avidLMQcManualAssessorData')
                        put_xnat(xc['usrpwd'], xnatAsses+'?avid:avidLMQcManualAssessorData/rater=Pawel')
                        put_xnat(xc['usrpwd'], xnatAsses+'?avid:avidLMQcManualAssessorData/startTime='+str(ttags[0]/1000.))
                        put_xnat(xc['usrpwd'], xnatAsses+'?avid:avidLMQcManualAssessorData/endTime='+str(ttags[1]/1000.))
                        put_xnat(xc['usrpwd'], xnatAsses+'?avid:avidLMQcManualAssessorData/duration='+str((ttags[1]-ttags[0])/1000.) )
                    except ValueError as ve:
                        print ' '
                        print '=============================================================='
                        print 'e> value error:', ve
                        print '=============================================================='
                        print ' '
                    else:
                        print '--------'
                        print 'i> QC attributes uploaded to XNAT.'
                        print '--------'

                    #sent the qc files to desired destination (qcpth + xnat)
                    xnatChart = xc['sbjs']+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/resources/CHARTS'
                    put_xnat(xc['usrpwd'], xnatChart+'?xsi:type=xnat:resourceCatalog&format=PNG')
                    xnatVideo = xc['sbjs']+'/' +sbjix+ '/experiments/' + eid[0]['ID'] + '/resources/VIDEOS'
                    put_xnat(xc['usrpwd'], xnatVideo+'?xsi:type=xnat:resourceCatalog&format=MP4')

                create_dir(qcfiles)

                for fn in os.listdir(outpth):
                    #------------------------------------------------
                    #copy the media files to dropobox or xnat
                    if fn.endswith(".png") or fn.endswith(".mp4"):
                        try:
                            shutil.copy( os.path.join(outpth,fn), os.path.join(qcfiles,fn) )
                        except IOError as e:
                            print ' '
                            print '=============================================================='
                            print 'IO error:', e
                            print '=============================================================='
                            print ' '
                        except shutil.Error as e:
                            print ' '
                            print '=============================================================='
                            print 'shutil error:', e
                            print '=============================================================='
                            print ' '
                        else:
                            print '--------'
                            print 'i> file:', os.path.join(outpth,fn), 'copied to:', os.path.join(qcfiles,fn)
                            print '--------'
                    #------------------------------------------------

                    if xnat_upload:
                        if fn==qccharts[0][0]:
                            put_xnatFile(xc['usrpwd'], xnatChart+'/files?content='+qccharts[1][0], os.path.join(outpth, qccharts[0][0]) )
                        elif fn==qccharts[0][1]:
                            put_xnatFile(xc['usrpwd'], xnatChart+'/files?content='+qccharts[1][1], os.path.join(outpth, qccharts[0][1]) )
                        elif fn==qccharts[0][2]:
                            put_xnatFile(xc['usrpwd'], xnatChart+'/files?content='+qccharts[1][2], os.path.join(outpth, qccharts[0][2]) )
                        elif fn==qcvideos[0][0]:
                            put_xnatFile(xc['usrpwd'], xnatVideo+'/files?content='+qcvideos[1][0], os.path.join(outpth, qcvideos[0][0]) )
                        elif fn==qcvideos[0][1]:
                            put_xnatFile(xc['usrpwd'], xnatVideo+'/files?content='+qcvideos[1][1], os.path.join(outpth, qcvideos[0][1]) )

                #remove the downloaded LM file
                if xc['rmvlm']:   os.remove(os.path.join(lmpth, lmfiles[i]['Name']))
                # =================================================================================--------------------

                print ' '
                print 'i> finished processing LM for:', sbjid, sbjlb
                if xnat_upload:
                    plist.append([sbjid, sbjlb, idate, 'y', time_stamp(), 'n', time_stamp()])
                else:
                    plist.append([sbjid, sbjlb, idate, 'n', time_stamp(), 'n', time_stamp()])
                return hst

        
#----------------------------------------------------------------------------------------------------------

#==========================================================================================================
#==========================================================================================================



#!/usr/bin/env python
"""init the NiftyPET package"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
# ---------------------------------------------------------------------------------

import os
import sys
import platform

# if using conda put the resources in the folder with the environment name
if 'CONDA_DEFAULT_ENV' in os.environ:
    env = os.environ['CONDA_DEFAULT_ENV']
    print 'i> conda environment found:', env
else:
    env = ''
# create the path for the resources files according to the OS platform
if platform.system() in ['Linux', 'Darwin']:
    path_resources = os.path.join( os.path.join(os.path.expanduser('~'),   '.niftypet'), env )
elif platform.system() == 'Windows':
    path_resources = os.path.join( os.path.join(os.getenv('LOCALAPPDATA'), '.niftypet'), env )
else:
    print 'e> unrecognised operating system!'
    
sys.path.append(path_resources)
try:
    import resources
except ImportError as ie:
    print '----------------------------'
    print 'e> Import Error: NiftyPET''s resources file <resources.py> could not be imported.  It should be in ''~/.niftypet/resources.py'' (Linux) or ''//Users//USERNAME//AppData//Local//niftypet//resources.py'' (Windows) but likely it does not exists.'
    print '----------------------------'
    raise ImportError
#===========================


import mmraux
from mmraux import explore_input as classify_input
from mmraux import mMR_params as get_mmrparams

# shared CUDA C library for extended auxiliary functions for the mMR
import mmr_auxe

import mmrnorm
import lm
import prj
import sct
import img

from dinf import gpuinfo, dev_info

from img.mmrimg import hdw_mumap
from img.mmrimg import obj_mumap
from img.mmrimg import pct_mumap
from img.mmrimg import align_mumap
from img.mmrimg import convert2e7 as im_dev2e7
from img.mmrimg import convert2dev as im_e72dev

from lm.mmrhist import dynamic_timings

from lm.mmrhist import mmrhist
from img.pipe import mmrchain

from prj.mmrprj import frwd_prj
from prj.mmrprj import back_prj

from prj.mmrsim import simulate_sino, simulate_recon

if resources.ENBLAGG:
    from lm.pviews import video_frm, video_dyn

if resources.ENBLXNAT:
    from xnat import xnat

# import sigaux
# import lms


"""Forward and back projector for PET data reconstruction"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
#------------------------------------------------------------------------------
import numpy as np
import sys
import os

import petprj

from niftypet.nipet.img import mmrimg
from niftypet.nipet import mmraux

#=========================================================================
# forward projector
#-------------------------------------------------------------------------
def frwd_prj(im, scanner_params, isub=np.array([-1], dtype=np.int32), dev_out=False, attenuation=False):
    ''' Calculate forward projection (a set of sinograms) for the provided input image.
        Arguments:
        im -- input image (can be emission or mu-map image).
        scanner_params -- dictionary of all scanner parameters, containing scanner constants,
            transaxial and axial look up tables (LUT).
        isub -- array of transaxial indices of all sinograms (angles x bins) used for subsets.
            when the first element is negative, all transaxial bins are used (as in pure EM-ML).
        dev_out -- if True, output sinogram is in the device form, i.e., with two dimensions
            (# bins/angles, # sinograms) instead of default three (# sinograms, # bins, # angles).
        attenuation -- controls whether emission or LOR attenuation probability sinogram
            is calculated; the default is False, meaning emission sinogram; for attenuation
            calculations (attenuation=True), the exponential of the negative of the integrated
            mu-values along LOR path is taken at the end.
    '''

    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    #>choose between attenuation forward projection (mu-map is the input)
    #>or the default for emission image forward projection
    if attenuation:
        att = 1
    else:
        att = 0

    if Cnt['SPN']==1:
        # number of rings calculated for the given ring range (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent (don't use ring range (1,63) as for this case no correction)
        if NRNG_c==64:
            nsinos -= 12
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']

    if im.shape[0]==Cnt['SO_IMZ'] and im.shape[1]==Cnt['SO_IMY'] and im.shape[2]==Cnt['SO_IMX']:
        ims = mmrimg.convert2dev(im, Cnt)
    elif im.shape[0]==Cnt['SZ_IMX'] and im.shape[1]==Cnt['SZ_IMY'] and im.shape[2]==Cnt['SZ_IMZ']:
        ims = im
    elif im.shape[0]==Cnt['rSO_IMZ'] and im.shape[1]==Cnt['SO_IMY'] and im.shape[2]==Cnt['SO_IMX']:
        ims = mmrimg.convert2dev(im, Cnt)
    elif im.shape[0]==Cnt['SZ_IMX'] and im.shape[1]==Cnt['SZ_IMY'] and im.shape[2]==Cnt['rSZ_IMZ']:
        ims = im
    else:
        print 'e> wrong image size;  it has to be one of these: (z,y,x) = (127,344,344) or (y,x,z) = (320,320,128)'

    if Cnt['VERBOSE']: print 'i> number of sinos:', nsinos
    
    #predefine the sinogram.  if subsets are used then only preallocate those bins which will be used.
    if isub[0]<0:
        sinog = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)
    else:
        sinog = np.zeros((len(isub), nsinos), dtype=np.float32)
    
    # --------------------
    petprj.fprj(sinog, ims, txLUT, axLUT, isub, Cnt, att)
    # --------------------
    # get the sinogram bins in a proper sinogram
    sino = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)
    if isub[0]>=0:    sino[isub,:] = sinog
    else:  sino = sinog

    # put the gaps back to form displayable sinogram
    if not dev_out:
        sino = mmraux.putgaps(sino, txLUT, Cnt)
    
    return sino

#=========================================================================
# back projector
#-------------------------------------------------------------------------
def back_prj(sino, scanner_params, isub=np.array([-1], dtype=np.int32)):
    ''' Calculate forward projection for the provided input image.
        Arguments:
        sino -- input emission sinogram to be back projected to the image space.
        scanner_params -- dictionary of all scanner parameters, containing scanner constants,
            transaxial and axial look up tables (LUT).
        isub -- array of transaxial indices of all sinograms (angles x bins) used for subsets;
            when the first element is negative, all transaxial bins are used (as in pure EM-ML).
    '''

    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    if Cnt['SPN']==1:
        # number of rings calculated for the given ring range (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent (don't use ring range (1,63) as for this case no correction)
        if NRNG_c==64:
            nsinos -= 12
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']


    #> check first the Siemens default sinogram;
    #> for this default shape only full sinograms are expected--no subsets.
    if len(sino.shape)==3:
        if sino.shape[0]!=nsinos or sino.shape[1]!=Cnt['NSANGLES'] or sino.shape[2]!=Cnt['NSBINS']:
            raise ValueError('Unexpected sinogram array dimensions/shape for Siemens defaults.')
        sinog = mmraux.remgaps(sino, txLUT, Cnt)

    elif len(sino.shape)==2:
        if isub[0]<0 and sino.shape[0]!=txLUT["Naw"]:
            raise ValueError('Unexpected number of transaxial elements in the full sinogram.')
        elif isub[0]>=0 and sino.shape[0]!=len(isub):
            raise ValueError('Unexpected number of transaxial elements in the subset sinogram.')
        #> check if the number of sinograms is correct
        if sino.shape[1]!=nsinos:
            raise ValueError('Inconsistent number of sinograms in the array.')
        #> when found the dimensions/shape are fine:
        sinog = sino
    else:
        raise ValueError('Unexpected shape of the input sinogram.')

    #predefine the output image depending on the number of rings used
    if Cnt['SPN']==1 and 'rSZ_IMZ' in Cnt:
        nvz = Cnt['rSZ_IMZ']
    else:
        nvz = Cnt['SZ_IMZ']
    bimg = np.zeros((Cnt['SZ_IMX'], Cnt['SZ_IMY'], nvz), dtype=np.float32)

    #> run back-projection
    petprj.bprj(bimg, sinog, txLUT, axLUT, isub, Cnt)

    #> change from GPU optimised image dimensions to the standard Siemens shape
    bimg = mmrimg.convert2e7(bimg, Cnt)

    return bimg
#-------------------------------------------------------------------------


"""Voxel-driven scatter modelling for PET data"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
#-------------------------------------------------------------------------------

import numpy as np
# import matplotlib.pyplot as plt
from math import pi
import random
import sys
import os
import scipy.ndimage as ndi

import nibabel as nib
import re

import petsct

from niftypet.nipet.prj import mmrprj, petprj, mmrrec
from niftypet.nipet.img import mmrimg
from niftypet.nipet import mmrnorm
from niftypet.nipet import mmraux
from niftypet.nipet import mmr_auxe


# ------------------------------------
def fwhm2sig (fwhm, Cnt):
    return (fwhm/Cnt['SO_VXY']) / (2*(2*np.log(2))**.5)
# ------------------------------------

#=================================================================================================
# S C A T T E R
#-------------------------------------------------------------------------------------------------

#get Klein-Nishina LUTs
def get_knlut(Cnt):
    from scipy.special import erfc

    SIG511 = Cnt['ER']*Cnt['E511']/2.35482

    CRSSavg = (2*(4/3.0-np.log(3)) + .5*np.log(3)-4/9.0)

    knlut = np.zeros((Cnt['NCOS'],2), dtype = np.float32)

    for i in range(Cnt['NCOS']):
        cosups = (Cnt['COSUPSMX']+i*Cnt['COSSTP'])
        alpha = 1/(2 - cosups)
        KNtmp = ( (0.5*Cnt['R02']) * alpha*alpha * ( alpha + 1/alpha - (1-cosups*cosups) ) )
        knlut[i,0] = KNtmp / ( 2*pi*Cnt['R02'] * CRSSavg);
        knlut[i,1] = ( (1+alpha)/(alpha*alpha)*(2*(1+alpha)/(1+2*alpha)-1/alpha*np.log(1+2*alpha)) + \
                        np.log(1+2*alpha)/(2*alpha)-(1+3*alpha)/((1+2*alpha)*(1+2*alpha)) ) / CRSSavg

        # Add energy resolution:
        if Cnt['ER']>0:
            print 'i> using energy resolution for scatter simulation, ER =', Cnt['ER']
            knlut[i,0] *= .5*erfc( (Cnt['LLD']-alpha*Cnt['E511'])/(SIG511*np.sqrt(2*alpha)) );
            #knlut[i,0] *= .5*erfc( (Cnt['LLD']-alpha*Cnt['E511'])/(SIG511) );

        # for large angles (small cosups) when the angle in GPU calculations is greater than COSUPSMX
        if (i==0):
            knlut[0,0] = 0;

    return knlut

#==================================================================================================
# GET SCATTER LUTs
#--------------------------------------------------------------------------------------------------
def rd2sni(offseg, r1, r0):
    rd = np.abs(r1-r0)
    rdi = (2*rd - 1*(r1>r0))
    sni = offseg[rdi] + np.minimum(r0,r1)
    return sni
#--------------------------------------------------------------------------------------------------
def get_sctLUT(Cnt):

    # get the indeces of rings used for scatter estimation
    irng = Cnt['SCTRNG']

    # get number of ring accounting for the possible ring reduction (to save computation time)
    # NRNG = Cnt['RNG_END']-Cnt['RNG_STRT']

    #-span-1 LUT (rings to sino index)
    seg = np.append( [Cnt['NRNG']], np.ceil( np.arange(Cnt['NRNG']-1,0,-.5) ).astype(np.int16) )
    offseg = np.int16( np.append( [0], np.cumsum(seg)) )

    #-3D scatter sino LUT. axial component based on michelogram.
    sctaxR = np.zeros((Cnt['NRNG']**2, 4), dtype=np.int32)
    sctaxW = np.zeros((Cnt['NRNG']**2, 4), dtype=np.float32)

    #-just for local check and display of the interpolation at work
    mich  = np.zeros((Cnt['NRNG'], Cnt['NRNG']), dtype=np.float32)
    mich2 = np.zeros((Cnt['NRNG'], Cnt['NRNG']), dtype=np.float32)


    J, I =  np.meshgrid(irng, irng)
    mich[J,I] = np.reshape(np.arange(len(Cnt['SCTRNG'])**2), (len(Cnt['SCTRNG']), len(Cnt['SCTRNG'])))
    # plt.figure(64), plt.imshow(mich, interpolation='none')

    for r1 in range(Cnt['RNG_STRT'], Cnt['RNG_END']):
        #border up and down
        bd = next(idx for idx in irng        if idx>=r1)
        bu = next(idx for idx in irng[::-1]  if idx<=r1)
        for r0 in range(Cnt['RNG_STRT'], Cnt['RNG_END']):

            # if (np.abs(r1-r0)>MRD):
            #     continue
            #border left and right
            br = next(idx for idx in irng        if idx>=r0)
            bl = next(idx for idx in irng[::-1]  if idx<=r0)
            #print '(r0,r1)=', r0,r1, '(bl,br,bu,bd)', bl,br,bu,bd

            #span-1 sino index (sni) creation:
            sni = rd2sni(offseg, r1, r0)

            #see: https://en.wikipedia.org/wiki/Bilinear_interpolation
            if (br==bl)and(bu!=bd):
                
                sctaxR[sni,0] = rd2sni(offseg, bd, r0)
                sctaxW[sni,0] = (r1-bu)/float(bd-bu)
                sctaxR[sni,1] = rd2sni(offseg, bu, r0)
                sctaxW[sni,1] = (bd-r1)/float(bd-bu)

                mich2[r1,r0] = mich[bd,r0]*sctaxW[sni,0]  +  mich[bu,r0]*sctaxW[sni,1] 

            elif (bu==bd)and(br!=bl):
                
                sctaxR[sni,0] = rd2sni(offseg, r1, bl)
                sctaxW[sni,0] = (br-r0)/float(br-bl)
                sctaxR[sni,1] = rd2sni(offseg, r1, br)
                sctaxW[sni,1] = (r0-bl)/float(br-bl)

                mich2[r1,r0] =  mich[r1,bl]*sctaxW[sni,0] + mich[r1,br]*sctaxW[sni,1]

            elif (bu==bd)and(br==bl):

                mich2[r1,r0] = mich[r1,r0]
                sctaxR[sni,0] = rd2sni(offseg, r1, r0)
                sctaxW[sni,0] = 1
                continue

            else:

                cf = float(((br-bl)*(bd-bu)))

                sctaxR[sni,0] = rd2sni(offseg, bd, bl)
                sctaxW[sni,0] = (br-r0)*(r1-bu)/cf
                sctaxR[sni,1] = rd2sni(offseg, bd, br)
                sctaxW[sni,1] = (r0-bl)*(r1-bu)/cf

                sctaxR[sni,2] = rd2sni(offseg, bu, bl)
                sctaxW[sni,2] = (br-r0)*(bd-r1)/cf
                sctaxR[sni,3] = rd2sni(offseg, bu, br)
                sctaxW[sni,3] = (r0-bl)*(bd-r1)/cf

                mich2[r1,r0] =  mich[bd,bl]*sctaxW[sni,0]+ mich[bd,br]*sctaxW[sni,1] + mich[bu,bl]*sctaxW[sni,2] + mich[bu,br]*sctaxW[sni,3]

    # plt.figure(65), plt.imshow(mich2, interpolation='none')

    #get K-N LUT:
    KN = get_knlut(Cnt)

    sctLUT = {'sctaxR':sctaxR, 'sctaxW':sctaxW, 'isrng':irng, 'offseg':offseg, 'KN':KN, 'mich_chck':[mich, mich2]} 
    return sctLUT


#-------------------------------------------------------------------------------------------------
# S C A T T E R    I N T E R P O L A T I O N
#-------------------------------------------------------------------------------------------------
def get_sctinterp(ssn2d, sct2AW, Cnt):

    from scipy.interpolate import griddata #used for scatter interpolation

    iAW = np.unique(sct2AW)
    nsbins = len(iAW)

    ai = iAW/Cnt['NSBINS']
    wi = iAW - ai*Cnt['NSBINS']

    #---add extra points (from the top) at the bottom of sino
    #find the indexes where angle indx is 0
    ai0 = np.where(ai==0)
    wie = abs((Cnt['NSBINS']-1)-wi[ai0])
    wien = len(wie)
    aie = Cnt['NSANGLES']*np.ones(wien, dtype=np.int32)

    #---prepare the arrays for interpolation input
    # indexes
    sind = np.zeros( (wien + nsbins,2), dtype=np.int32)
    # scatter values for the above indecies
    sval = np.zeros( (wien + nsbins,1), dtype=np.float32)
    # append to the sino bottom in order to get a properly interpolated sino
    sind[:,1] = np.append(wi,wie)
    sind[:,0] = np.append(ai,aie)
    sval = ssn2d[ np.append( ai, np.zeros(len(ai0[0]),dtype=np.int32) ), np.append(wi, wi[ai0]) ]

    # do the interpolation after creating the grid
    grid_x, grid_y = np.mgrid[0:Cnt['NSANGLES'], 0:Cnt['NSBINS']]
    issino = griddata(sind, sval, (grid_x, grid_y), method='cubic')
    issino = np.nan_to_num(issino)
    issino[issino<0] = 0

    return issino
#====================================================================================================


def vsm(mumaps, em, datain, hst, rsinos, scanner_params, prcntScl=0.1, emmsk=False):

    muh, muo = mumaps


    #-constants, transaxial and axial LUTs are extracted
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    if emmsk and not os.path.isfile(datain['em_nocrr']):
        if Cnt['VERBOSE']: print 'i> reconstruction of emission data without scatter and attenuation correction for mask generation'
        recnac = mmrrec.osemone(datain, mumaps, hst, scanner_params, recmod=0, itr=3, fwhm=2.0, store_img=True)
        datain['em_nocrr'] = recnac.fpet


    #-get the normalisation components
    nrmcmp, nhdr = mmrnorm.get_components(datain, Cnt)

    #-smooth for defining the sino scatter only regions
    mu_sctonly =  ndi.filters.gaussian_filter(
        mmrimg.convert2dev(muo, Cnt),
        fwhm2sig(0.42, Cnt),
        mode='mirror'
    )

    if Cnt['SPN']==1:
        snno = Cnt['NSN1']
        snno_= Cnt['NSN64']
        ssrlut = axLUT['sn1_ssrb']
        saxnrm = nrmcmp['sax_f1']
    elif Cnt['SPN']==11:
        snno = Cnt['NSN11']
        snno_= snno
        ssrlut = axLUT['sn11_ssrb']
        saxnrm = nrmcmp['sax_f11']

    #LUTs for scatter
    sctLUT = get_sctLUT(Cnt)
    
    
    #-smooth before down-sampling mu-map and emission image
    muim = ndi.filters.gaussian_filter(muo+muh, fwhm2sig(0.42, Cnt), mode='mirror')
    muim = ndi.interpolation.zoom( muim, Cnt['SCTSCLMU'], order=3 ) #(0.499, 0.5, 0.5)

    emim = ndi.filters.gaussian_filter(em, fwhm2sig(0.42, Cnt), mode='mirror')
    emim = ndi.interpolation.zoom( emim, Cnt['SCTSCLEM'], order=3 ) #(0.34, 0.33, 0.33)
    #emim = ndi.interpolation.zoom( emim, (0.499, 0.5, 0.5), order=3 )
    

    #-smooth the mu-map for mask creation.  the mask contains voxels for which attenuation ray LUT is found.
    smomu = ndi.filters.gaussian_filter(muim, fwhm2sig(0.84, Cnt), mode='mirror')
    mumsk = np.int8(smomu>0.003)

    #CORE SCATTER ESTIMATION
    NSCRS, NSRNG = 64, 8
    sctout ={
        'xsxu'    :np.zeros((NSCRS, NSCRS/2), dtype=np.int8), #one when xs>xu, otherwise zero
        'bin_indx':np.zeros((NSCRS, NSCRS/2), dtype=np.int32),
        'sct_val' :np.zeros((Cnt['TOFBINN'], NSRNG, NSCRS, NSRNG, NSCRS/2), dtype=np.float32),
        'sct_3d'  :np.zeros((Cnt['TOFBINN'], snno_, NSCRS, NSCRS/2), dtype=np.float32)
    }

    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>
    petsct.scatter(sctout, muim, mumsk, emim, sctLUT, txLUT, axLUT, Cnt)
    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>
    sct3d  = sctout['sct_3d']
    sctind = sctout['bin_indx']

    if Cnt['VERBOSE']: print 'i> total scatter sum:', np.sum(sct3d)
    if np.sum(sct3d)<1e-04:
        sss    = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
        amsksn = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
        sssr   = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
        return sss, sssr, amsksn

    #get SSR for randoms from span-1 or span-11
    rssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
    for i in range(snno):
        rssr[ssrlut[i],:,:] += rsinos[i,:,:]

    #ATTENUATION FRACTIONS for scatter only regions, and NORMALISATION for all SCATTER
    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>
    currentspan = Cnt['SPN']
    Cnt['SPN'] = 1
    atto = np.zeros((txLUT['Naw'], Cnt['NSN1']), dtype=np.float32)
    petprj.fprj(atto, mu_sctonly, txLUT, axLUT, np.array([-1], dtype=np.int32), Cnt, 1)
    atto = mmraux.putgaps(atto, txLUT, Cnt)
    #--------------------------------------------------------------
    #get norm components setting the geometry and axial to ones as they are accounted for differently
    nrmcmp['geo'][:] = 1
    nrmcmp['axe1'][:] = 1
    #get sino with no gaps
    nrmg = np.zeros((txLUT['Naw'], Cnt['NSN1']), dtype=np.float32)
    mmr_auxe.norm(nrmg, nrmcmp, hst['buckets'], axLUT, txLUT['aw2ali'], Cnt)
    nrm = mmraux.putgaps(nrmg, txLUT, Cnt)
    #--------------------------------------------------------------
    

    #get attenuation + norm in (span-11) and SSR
    attossr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
    nrmsssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
    for i in range(Cnt['NSN1']):
        si = axLUT['sn1_ssrb'][i]
        attossr[si,:,:] += atto[i,:,:] / float(axLUT['sn1_ssrno'][si])
        nrmsssr[si,:,:] += nrm[i,:,:] / float(axLUT['sn1_ssrno'][si])
    if currentspan==11:
        Cnt['SPN']=11
        nrmg = np.zeros((txLUT['Naw'], snno), dtype=np.float32)
        mmr_auxe.norm(nrmg, nrmcmp, hst['buckets'], axLUT, txLUT['aw2ali'], Cnt)
        nrm = mmraux.putgaps(nrmg, txLUT, Cnt)
    #--------------------------------------------------------------
    
    #get the mask for the object from uncorrected emission image
    if emmsk and os.path.isfile(datain['em_nocrr']):
        nim = nib.load(datain['em_nocrr'])
        A   = nim.get_sform()
        eim = np.float32( nim.get_data() )
        eim = eim[:,::-1,::-1]
        eim = np.transpose(eim, (2, 1, 0))

        em_sctonly = ndi.filters.gaussian_filter(eim, fwhm2sig(.6, Cnt), mode='mirror')
        msk = np.float32(em_sctonly>0.07*np.max(em_sctonly))
        msk = ndi.filters.gaussian_filter(msk, fwhm2sig(.6, Cnt), mode='mirror')
        msk = np.float32(msk>0.01)
        msksn = mmrprj.frwd_prj(msk, txLUT, axLUT, Cnt)

        mssr = mmraux.sino2ssr(msksn, axLUT, Cnt)
        mssr = mssr>0
    else:
        mssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.bool);

    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>

    #--------------------------------------------------------------------------------------------
    # get scatter sinos for TOF or non-TOF
    if Cnt['TOFBINN']>1:
        ssn = np.zeros((Cnt['TOFBINN'], snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float64);
        sssr = np.zeros((Cnt['TOFBINN'], Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
        tmp2d = np.zeros((Cnt['NSANGLES']*Cnt['NSBINS']), dtype=np.float64)
        if Cnt['VERBOSE']: print 'i> interpolate each scatter sino...'
        for k in range(Cnt['TOFBINN']):
            if Cnt['VERBOSE']: print 'i> doing TOF bin k =', k
            for i in range(snno):
                tmp2d[:] = 0
                for ti in range(len(sctind)):
                    tmp2d[ sctind[ti] ] += sct3d[k,i,ti]
                #interpolate estimated scatter
                ssn[k,i,:,:] = get_sctinterp( np.reshape(tmp2d, (Cnt['NSANGLES'], Cnt['NSBINS'])), sctind, Cnt )
                sssr[k, ssrlut[i], :, :] += ssn[k,i,:,:]
            if Cnt['VERBOSE']: print 'i> TOF bin #', k
    elif Cnt['TOFBINN']==1:
        ssn = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
        sssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
        tmp2d = np.zeros((Cnt['NSANGLES']*Cnt['NSBINS']), dtype=np.float32)
        if Cnt['VERBOSE']: print 'i> scatter sinogram interpolation...'
        for i in range(snno):
            tmp2d[:] = 0
            for ti in range(len(sctind)):
                tmp2d[ sctind[ti] ] += sct3d[0,i,ti]
            #interpolate estimated scatter
            ssn[i,:,:] = get_sctinterp( np.reshape(tmp2d, (Cnt['NSANGLES'], Cnt['NSBINS'])), sctind, Cnt )
            sssr[ssrlut[i],:,:] += ssn[i,:,:]
            if Cnt['VERBOSE'] and (i%100)==0: print 'i> ', i, 'sinograms interpolated'
    #--------------------------------------------------------------------------------------------

    #=== scale scatter for ssr and non-TOF===
    #mask
    rmsk = (txLUT['msino']>0).T
    rmsk.shape = (1,Cnt['NSANGLES'],Cnt['NSBINS'])
    rmsk = np.repeat(rmsk, Cnt['NSEG0'], axis=0)
    amsksn = np.logical_and( attossr>=0.999, rmsk) * ~mssr
    #scaling factors for ssr
    scl_ssr = np.zeros( (Cnt['NSEG0']), dtype=np.float32)
    for sni in range(Cnt['NSEG0']):
        # region of choice for scaling
        thrshld = prcntScl * np.max(sssr[sni,:,:])
        amsksn[sni,:,:] *= (sssr[sni,:,:]>thrshld)
        amsk = amsksn[sni,:,:]
        #normalised estimated scatter
        mssn = sssr[sni,:,:] * nrmsssr[sni,:,:]
        mssn[np.invert(amsk)] = 0
        #vectorised masked sino
        vssn = mssn[amsk]
        vpsn = hst['pssr'][sni, amsk] - rssr[sni, amsk]
        scl_ssr[sni] = np.sum(vpsn) / np.sum(mssn)
        #ssr output
        sssr[sni,:,:] *= nrmsssr[sni,:,:]*scl_ssr[sni]


    #=== scale scatter for the proper sino ===
    sss = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
    for i in range(snno):
        sss[i,:,:] = ssn[i,:,:]*scl_ssr[ssrlut[i]]*saxnrm[i] * nrm[i,:,:]

    return sss, sssr, amsksn





"""module for pipelined image reconstruction and analysis"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
#------------------------------------------------------------------------------

import numpy as np
import sys
import os
import scipy.ndimage as ndi
from subprocess import call

from niftypet import nimpa

from niftypet.nipet.lm import dynamic_timings
from niftypet.nipet.prj import mmrrec
from niftypet.nipet.img import obtain_image
from niftypet.nipet.img.mmrimg import image_affine
from niftypet.nipet.lm.mmrhist import mmrhist

integers = (int, np.int32, np.int16, np.int8, np.uint8, np.uint16, np.uint32)

#------------------------------------------------------------------------------
def mmrchain(datain,        # all input data in a dictionary
            scanner_params, # all scanner parameters in one dictionary
                            # containing constants, transaxial and axial
                            # LUTs.
            outpath='',     # output path for results
            frames=['fluid', [0,0]], # definition of time frames.
            mu_h = [],      # hardware mu-map.
            mu_o = [],      # object mu-map.
            tAffine = [],   # affine transformations for the mu-map for
                            # each time frame separately.

            itr=4,          # number of OSEM iterations
            fwhm=0.,        # Gaussian Smoothing FWHM
            recmod = -1,    # reconstruction mode: -1: undefined, chosen
                            # automatically. 3: attenuation and scatter
                            # correction, 1: attenuation correction
                            # only, 0: no correction (randoms only).
            histo=[],       # input histogram (from list-mode data);
                            # if not given, it will be performed.

            trim=False,
            trim_scale=2,
            trim_interp=1,  # interpolation for upsampling used in PVC
            trim_memlim=True,   # reduced use of memory for machines
                                # with limited memory (slow though)

            pvcroi=[],      # ROI used for PVC.  If undefined no PVC
                            # is performed.

            pvcreg_tool = 'nifyreg', # the registration tool used in PVC
            store_rois = False, # stores the image of PVC ROIs 
                                # as defined in pvcroi.

            psfkernel=[],
            pvcitr=5,

            fcomment='',    # text comment used in the file name of
                            # generated image files
            ret_sinos=False,# return prompt, scatter and randoms
                            # sinograms for each reconstruction
            store_img = True,
            store_img_intrmd=False,
            store_itr=[],   # store any reconstruction iteration in
                            # the list.  ignored if the list is empty.
            del_img_intrmd=False):


    # decompose all the scanner parameters and constants
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']


    # -------------------------------------------------------------------------
    # FRAMES
    # check for the provided dynamic frames
    if isinstance(frames, list):
        # Can be given in three ways:
        # * a 1D list (duration of each frame is listed)
        # * a more concise 2D list--repetition and duration lists in
        #   each entry.  Must start with the 'def' entry.
        # * a 2D list with fluid timings: must start with the string
        #   'fluid' or 'timings.  a 2D list with consecutive lists
        #   describing start and end of the time frame, [t0, t1];
        #   The number of time frames for this option is unlimited,
        #   provided the t0 and t1 are within the acquisition times.

        # 2D starting with entry 'fluid' or 'timings'
        if  isinstance(frames[0], basestring) and (frames[0]=='fluid' or frames[0]=='timings') \
            and all([isinstance(t,list) and len(t)==2 for t in frames[1:]]):
            t_frms = frames[1:]

        # if 2D definitions, starting with entry 'def':
        elif isinstance(frames[0], basestring) and frames[0]=='def' \
            and all([isinstance(t,list) and len(t)==2 for t in frames[1:]]):
            # get total time and list of all time frames
            dfrms = dynamic_timings(frames)
            t_frms = dfrms['timings'][1:]

        # if 1D:
        elif all([isinstance(t, integers) for t in frames]):
            # get total time and list of all time frames
            dfrms = dynamic_timings(frames)
            t_frms = dfrms['timings'][1:]

        else:
            print 'e> osemdyn: frames definitions are not given in the correct list format: 1D [15,15,30,30,...] or 2D list [[2,15], [2,30], ...]'
    else:
        print 'e> osemdyn: provided dynamic frames definitions are not in either Python list or nympy array.'
        raise TypeError('Wrong data type for dynamic frames')
    # number of dynamic time frames
    nfrm = len(t_frms)
    # -------------------------------------------------------------------------



    # -------------------------------------------------------------------------
    # create folders for results
    if outpath=='':
        petdir = os.path.join(datain['corepath'], 'reconstructed')
        fmudir = os.path.join(datain['corepath'], 'mumap-obj')
        pvcdir = os.path.join(datain['corepath'], 'PRCL')
    else:
        petdir = os.path.join(outpath, 'PET')
        fmudir = os.path.join(outpath, 'mumap-obj')
        pvcdir = os.path.join(outpath, 'PRCL')

    # folder for co-registered mu-maps (for motion compensation)
    fmureg = os.path.join( fmudir, 'registered')
    # folder for affine transformation MR/CT->PET
    petaff = os.path.join( petdir, 'faffine')

    # folder for reconstructed images (dynamic or static depending on number of frames).
    if nfrm>1:
        petimg = os.path.join(petdir, 'multiple-frames')
        pvcdir = os.path.join(pvcdir, 'multiple-frames')
    elif nfrm==1:
        petimg = os.path.join(petdir, 'single-frame')
        pvcdir = os.path.join(pvcdir, 'single-frame')
    else:
        print 'e> confused!'
        raise TypeError('Unrecognised time frames!')
    # create now the folder
    nimpa.create_dir(petimg)
    # create folder
    nimpa.create_dir(petdir)
    # -------------------------------------------------------------------------


    # -------------------------------------------------------------------------
    # MU-MAPS
    # get the mu-maps, if given;  otherwise will use blank mu-maps.
    if tAffine:
        muod = obtain_image(mu_o, imtype='object mu-map', verbose=Cnt['VERBOSE'])
    else:
        muod = obtain_image(mu_o, Cnt=Cnt, imtype='object mu-map')

    # hardware mu-map
    muhd = obtain_image(mu_h, Cnt, imtype='hardware mu-map')

    # choose the mode of reconstruction based on the provided (or not) mu-maps
    if muod['exists'] and muhd['exists'] and recmod==-1:
        recmod = 3
    elif  (muod['exists'] or muhd['exists']) and recmod==-1:
        recmod = 1
        print 'w> partial mu-map:  scatter correction is switched off.'
    else:
        if recmod==-1:
            recmod = 0
            print 'w> no mu-map provided: scatter and attenuation corrections are switched off.'
    # -------------------------------------------------------------------------

    #import pdb; pdb.set_trace()

    # output dictionary
    output = {}
    output['recmod'] = recmod
    output['frames'] = t_frms
    output['#frames'] = nfrm

    # if affine transformation is given the baseline mu-map in NIfTI file or dictionary has to be given
    if not tAffine:
        if Cnt['VERBOSE']: print 'i> using the provided mu-map the same way for all frames.'
    else:
        if len(tAffine)!=nfrm:
            print 'e> the number of affine transformations in the list has to be the same as the number of dynamic frames!'
            raise IndexError('Inconsistent number of frames.')
        elif not isinstance(tAffine, list):
            print 'e> tAffine has to be a list of either 4x4 numpy arrays of affine transformations or a list of file path strings!'
            raise IndexError('Expecting a list.')
        elif not 'fim' in muod:
            print 'e> when tAffine is given, the object mu-map has to be provided either as a dictionary or NIfTI file!'
            raise NameError('No path to object mu-map.')

        # check if all are file path strings to the existing files
        if all([isinstance(t, basestring) for t in tAffine]):
            if all([os.path.isfile(t) for t in tAffine]):
                # the internal list of affine transformations
                faff_frms = tAffine
                if Cnt['VERBOSE']: print 'i> using provided paths to affine transformations for each dynamic frame.'
            else:
                print 'e> not all provided paths are valid!'
                raise IOError('Wrong paths.')
        # check if all are numpy arrays
        elif all([isinstance(t, (np.ndarray, np.generic)) for t in tAffine]):
            # create the folder for dynamic affine transformations
            nimpa.create_dir(petaff)
            faff_frms = []
            for i in range(nfrm):
                fout = os.path.join(petaff, 'affine_frame('+str(i)+').txt')
                np.savetxt(fout, tAffine[i], fmt='%3.9f')
                faff_frms.append(fout)
            if Cnt['VERBOSE']: print 'i> using provided numpy arrays affine transformations for each dynamic frame.'
        else:
            raise StandardError('Affine transformations for each dynamic frame could not be established.')

        # -------------------------------------------------------------------------------------
        # get ref image for mu-map resampling
        # -------------------------------------------------------------------------------------
        if 'fmuref' in muod:
            fmuref = muod['fmuref']
            if Cnt['VERBOSE']: print 'i> reusing the reference mu-map from the object mu-map dictionary.'
        else:
            # create folder if doesn't exists
            nimpa.create_dir(fmudir)
            # ref file name
            fmuref = os.path.join(fmudir, 'muref.nii.gz')
            # ref affine
            B = image_affine(datain, Cnt, gantry_offset=False)
            # ref image (blank)
            im = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)
            # store ref image
            nimpa.array2nii(im, B, fmuref)
            if Cnt['VERBOSE']: print 'i> generated a reference mu-map in', fmuref
        # -------------------------------------------------------------------------------------

        output['fmuref'] = fmuref
        output['faffine'] = faff_frms

    # output list of intermidiate file names for mu-maps and PET images (useful for dynamic imaging)
    if tAffine: output['fmureg'] = []
    if store_img_intrmd: output['fpeti'] = []

    # dynamic images in one numpy array
    dynim = np.zeros((nfrm, Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMY']), dtype=np.float32)
    #if asked, output only scatter+randoms sinogram for each frame
    if ret_sinos and itr>1 and recmod>2:
        dynrsn = np.zeros((nfrm, Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        dynssn = np.zeros((nfrm, Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        dynpsn = np.zeros((nfrm, Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)


    # import pdb; pdb.set_trace()

    # starting frame index with reasonable prompt data
    ifrmP = 0
    # iterate over frame index
    for ifrm in range(nfrm):
        # start time of a current (ifrm-th) dynamic frame
        t0 = int(t_frms[ifrm][0])
        # end time of a current (ifrm-th) dynamic frame
        t1 = int(t_frms[ifrm][1])
        # --------------
        # check if there is enough prompt data to do a reconstruction
        # --------------
        print 'i> dynamic frame times t0, t1:', t0, t1
        if not histo:
            hst = mmrhist(datain, scanner_params, t0=t0, t1=t1)
        else:
            hst = histo
            print ''
            print 'i> using provided histogram'
            print ''
        if np.sum(hst['dhc'])>0.99*np.sum(hst['phc']):
            print '==============================================================================================='
            print 'w> the amount of random events is the greatest part of prompt events => omitting reconstruction'
            print '==============================================================================================='
            ifrmP = ifrm+1
            continue
        # --------------------
        # transform the mu-map if given the affine transformation for each frame
        if tAffine:
            # create the folder for aligned (registered for motion compensation) mu-maps
            nimpa.create_dir(fmureg)
            # the converted nii image resample to the reference size
            fmu = os.path.join(fmureg, 'mumap_dyn_frm'+str(ifrm)+fcomment+'.nii.gz')
            # command for resampling
            if os.path.isfile( Cnt['RESPATH'] ):
                cmd = [Cnt['RESPATH'],
                '-ref', fmuref,
                '-flo', muod['fim'],
                '-trans', faff_frms[ifrm],
                '-res', fmu,
                '-pad', '0']
                if not Cnt['VERBOSE']: cmd.append('-voff')
                call(cmd)
            else:
                print 'e> path to the executable for resampling is incorrect!'
                raise IOError('Incorrect NiftyReg (resampling) executable.')
            # get the new mu-map from the just resampled file
            muodct = nimpa.getnii(fmu, output='all')
            muo = muodct['im']
            A = muodct['affine']
            muo[muo<0] = 0
            output['fmureg'].append(fmu)
        else:
            muo = muod['im']
        #---------------------

        # output image file name
        if nfrm>1:
            frmno = '_frm'+str(ifrm)
        else:
            frmno = ''

        # run OSEM reconstruction of a single time frame
        recimg = mmrrec.osemone(datain, [muhd['im'], muo],
                                hst, scanner_params,
                                recmod=recmod, itr=itr, fwhm=fwhm,
                                outpath=petimg,
                                frmno=frmno,
                                fcomment=fcomment+'_i',
                                store_img=store_img_intrmd,
                                store_itr=store_itr,
                                ret_sinos=ret_sinos)
        # form dynamic numpy array
        dynim[ifrm,:,:,:] = recimg.im
        if ret_sinos and itr>1 and recmod>2:
            dynpsn[ifrm,:,:,:] = hst['psino']
            dynssn[ifrm,:,:,:] = recimg.ssn
            dynrsn[ifrm,:,:,:] = recimg.rsn

        if store_img_intrmd: output['fpeti'].append(recimg.fpet)
        if nfrm==1: output['tuple'] = recimg

    output['im'] = np.squeeze(dynim)
    if ret_sinos and itr>1 and recmod>2:
        output['sinos'] = {'psino':dynpsn, 'ssino':dynssn, 'rsino':dynrsn}

    # ----------------------------------------------------------------------
    # trim the PET image
    # images have to be stored for PVC
    if pvcroi: store_img_intrmd = True
    if trim:
        # create file name
        if 'lm_dcm' in datain:
            fnm = os.path.basename(datain['lm_dcm'])[:20]
        elif 'lm_ima' in datain:
            fnm = os.path.basename(datain['lm_ima'])[:20]
        # trim PET and upsample
        petu = nimpa.trimim(
            dynim,
            affine=image_affine(datain, Cnt),
            scale=trim_scale,
            int_order=trim_interp,
            outpath=petimg,
            fname = fnm,
            fcomment=fcomment,
            store_img_intrmd=store_img_intrmd,
            memlim=trim_memlim,
            verbose=Cnt['VERBOSE']
        )

        output.update({'trimmed': { 'im':petu['im'],
                                    'fpet':petu['fimi'],
                                    'affine':petu['affine']}}
        )
    # ----------------------------------------------------------------------


    # ----------------------------------------------------------------------
    #run PVC if requested and required input given
    if pvcroi:
        if not os.path.isfile(datain['T1lbl']):
            print 'e> no label image from T1 parcellations and/or ROI definitions!'
            raise StandardError('No ROIs')
        else:
            # get the PSF kernel for PVC
            if not psfkernel:
                psfkernel = nimpa.psf_measured(scanner='mmr', scale=trim_scale)
            else:
                if isinstance(psfkernel, (np.ndarray, np.generic)) and psfkernel.shape!=(3, 17):
                    print 'e> the PSF kernel has to be an numpy array with the shape of (3, 17)!'
                    raise IndexError('PSF: wrong shape or not a matrix')
        
        #> file names for NIfTI images of PVC ROIs and PVC corrected PET
        froi = []
        fpvc = []

        #> perform PVC for each time frame
        dynpvc = np.zeros(petu['im'].shape, dtype=np.float32)
        for i in range(ifrmP,nfrm):
            # transform the parcellations (ROIs) if given the affine transformation for each frame
            if not tAffine:
                print 'w> affine transformation are not provided: will generate for the time frame.'
                faffpvc = ''
                #raise StandardError('No affine transformation')
            else:
                faffpvc = faff_frms[i]
            # chose file name of individual PVC images
            if nfrm>1:
                fcomment_pvc = '_frm'+str(i)+fcomment
            else:
                fcomment_pvc = fcomment
            #============================
            # perform PVC
            petpvc_dic = nimpa.pvc_iyang(
                petu['fimi'][i],
                datain,
                Cnt,
                pvcroi,
                psfkernel,
                tool = pvcreg_tool,
                itr = pvcitr,
                faff = faffpvc,
                fcomment = fcomment_pvc,
                outpath = pvcdir,
                store_rois = store_rois,
                store_img = store_img_intrmd)
            #============================
            if nfrm>1:
                dynpvc[i,:,:,:] = petpvc_dic['im']
            else:
                dynpvc = petpvc_dic['im']
            
            fpvc.append(petpvc_dic['fpet'])

            if store_rois: froi.append(petpvc_dic['froi'])
        
        #> update output dictionary
        output.update({'impvc':dynpvc})
        if store_img_intrmd: output.update({'fpvc':fpvc})
        if store_rois: output.update({'froi':froi})
    # ----------------------------------------------------------------------

    if store_img:
        # description for saving NIFTI image
        # attenuation number: if only bed present then it is 0.5
        attnum =  ( 1*muhd['exists'] + 1*muod['exists'] ) / 2.
        descrip =    'alg=osem'                     \
                    +';att='+str(attnum*(recmod>0)) \
                    +';sct='+str(1*(recmod>1))      \
                    +';spn='+str(Cnt['SPN'])        \
                    +';sub=14'                      \
                    +';itr='+str(itr)               \
                    +';fwhm='+str(fwhm)             \
                    +';nfrm='+str(nfrm)

        # squeeze the not needed dimensions
        dynim = np.squeeze(dynim)

        # NIfTI file name for the full PET image (single or multiple frame)

        # save the image to NIfTI file
        if nfrm==1:
            t0 = hst['t0']
            t1 = hst['t1']
            if t1==t0:
                t0 = 0
                t1 = hst['dur']
            fpet = os.path.join(
                    petimg,
                    os.path.basename(recimg.fpet)[:8] \
                    +'_t-'+str(t0)+'-'+str(t1)+'sec' \
                    +'_itr-'+str(itr) )
            fpeto = fpet+fcomment+'.nii.gz'
            nimpa.prc.array2nii( dynim[::-1,::-1,:], recimg.affine, fpeto, descrip=descrip)
        else:
            fpet = os.path.join(
                    petimg,
                    os.path.basename(recimg.fpet)[:8]\
                    +'_nfrm-'+str(nfrm)+'_itr-'+str(itr)
                )
            fpeto = fpet+fcomment+'.nii.gz'
            nimpa.prc.array2nii( dynim[:,::-1,::-1,:], recimg.affine, fpeto, descrip=descrip)

        # get output file names for trimmed/PVC images
        if trim:
            # folder for trimmed and dynamic
            pettrim = os.path.join( petimg, 'trimmed')
            # make folder
            nimpa.create_dir(pettrim)
            # trimming scale added to NIfTI descritoption
            descrip_trim = descrip + ';trim_scale='+str(trim_scale)
            # file name for saving the trimmed image
            fpetu = os.path.join(pettrim, os.path.basename(fpet) + '_trimmed-upsampled-scale-'+str(trim_scale))
            # in case of PVC
            if pvcroi:
                # itertive Yang (iY) added to NIfTI descritoption
                descrip_pvc = descrip_trim + ';pvc=iY'
                # file name for saving the PVC NIfTI image
                fpvc = fpetu + '_PVC' + fcomment + '.nii.gz'
                output['trimmed']['fpvc'] = fpvc

            # update the trimmed image file name
            fpetu += fcomment+'.nii.gz'
            # store the file name in the output dictionary
            output['trimmed']['fpet'] = fpetu

        output['fpet'] = fpeto

        # save images
        if nfrm==1:
            if trim:
                nimpa.prc.array2nii( petu['im'][::-1,::-1,:], petu['affine'], fpetu, descrip=descrip_trim)
            if pvcroi:
                nimpa.prc.array2nii( dynpvc[::-1,::-1,:], petu['affine'], fpvc, descrip=descrip_pvc)
        elif nfrm>1:
            if trim:
                nimpa.prc.array2nii( petu['im'][:,::-1,::-1,:], petu['affine'], fpetu, descrip=descrip_trim)
            if pvcroi:
                nimpa.prc.array2nii( dynpvc[:,::-1,::-1,:], petu['affine'], fpvc, descrip=descrip_pvc)


    if del_img_intrmd:
        if pvcroi:
            for fi in fpvc:
                os.remove(fi)
        if trim:
            for fi in petu['fimi']:
                os.remove(fi)




    return output



#!/usr/bin/python
__author__ = 'pawel'

import numpy as np
import os
import sys

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
import matplotlib.animation as manimation

def mvavg(interval, window_size):
    window = np.ones(int(window_size))/float(window_size)
    return np.convolve(interval, window, 'same')

def video_frm(hst, outpth):

    plt.close('all')

    #=============== CONSTANTS ==================
    VTIME = 4
    #============================================

    i = np.argmax(hst['phc'])
    ymin = np.floor( min(hst['cmass'][i:i+300]) )
    ymax = np.ceil( max(hst['cmass'][i+100:]) )

    mfrm = hst['pvs_sgtl'].shape[0];

    #--for movie
    FFMpegWriter = manimation.writers['ffmpeg']
    metadata = dict(title='GPU Sino Views', artist='Pawel', comment=':)')
    writer = FFMpegWriter(fps=25, bitrate=30000, metadata=metadata)
    #--

    fig3 = plt.figure()

    ax1 = plt.subplot(311)
    plt.title('Coronal View')
    plt.setp( ax1.get_xticklabels(), visible=False)
    plt.tick_params(axis='both', which='both', bottom='off', top='off', labelbottom='off')
    l1 = plt.imshow(hst['pvs_crnl'][100,:,:]/np.mean(hst['pvs_crnl'][100,:,:]), cmap='jet',interpolation='nearest')

    ax2 = plt.subplot(312)
    plt.title('Sagittal View')
    plt.setp( ax2.get_xticklabels(), visible=False)
    plt.tick_params(axis='both', which='both', bottom='off', top='off', labelbottom='off')
    l = plt.imshow(hst['pvs_sgtl'][100,:,:]/np.mean(hst['pvs_sgtl'][100,:,:]), cmap='jet',interpolation='nearest')

    ax3 = plt.subplot(313)
    plt.title('Axial Centre of Mass')
    t = np.arange(0., hst['dur'], 1.)
    #plt.plot(t, rprmt, 'k', t, rdlyd, 'r')
    plt.plot(t, mvavg(hst['cmass'][:],5),'k')
    plt.ylim([ymin, ymax])
    plt.xlabel('Time [s]')
    l2, = plt.plot(np.array([1000, 1000]), np.array([0, ymax]), 'b')

    #how many gpu frames per movie (controls the time resolution)
    mf = 6
    mmfrm = mfrm/mf

    fnm = os.path.join(outpth, 'pViews_' +str(mf)+'.mp4')

    with writer.saving( fig3, fnm, 200 ):
        for i in range(mmfrm):
            print('i> short frame to movie:', i)
            tmp = np.sum( hst['pvs_sgtl'][mf*i:mf*(i+1),:,:], axis=0)
            tmp2= np.sum( hst['pvs_crnl'][mf*i:mf*(i+1),:,:], axis=0)
            tmp = tmp/np.mean(tmp)
            tmp2 = tmp2/np.mean(tmp2)
            l.set_data(tmp)
            l1.set_data(tmp2)
            # l2.set_data(VTIME*mf*i*np.ones(2), np.array([0, np.max(hst['phc'])]))
            l2.set_data(VTIME*mf*i*np.ones(2), np.array([0, ymax]))
            writer.grab_frame()


    plt.show()
    return fnm

#===================================================================================
# Dynamic Frames to Projection Views
#-----------------------------------------------------------------------------------
def video_dyn(hst, frms, outpth, axLUT, Cnt):

    plt.close('all')

    #=============== CONSTANTS ==================
    VTIME = 4
    NRINGS = Cnt['NRNG']
    NSN11 = Cnt['NSN11']
    NDSN = Cnt['NSEG0']
    A = Cnt['NSANGLES']
    W = Cnt['NSBINS']

    voxz = Cnt['SO_VXZ']
    nsinos = NSN11
    #============================================

    # for scaling of the mass centre
    i = np.argmax(hst['phc'])
    ymin = np.floor( min(hst['cmass'][i:i+300]) )
    ymax = np.ceil( max(hst['cmass'][i+100:]) )

    # number of dynamic frames
    nfrm = hst['psino'].shape[0]
    # and cumulative sum of frame duration
    frmcum = np.cumsum(frms)

    # dynamic sino
    ddsino = np.zeros((nfrm, NDSN, A, W), dtype=np.int32)
    gsum = np.zeros(nfrm, dtype=np.int32)
    gpu_totsum = 0

    for frm in range(nfrm):
        for i in range(nsinos):
            ddsino[frm, axLUT['sn11_ssrb'][i], :, :] += hst['psino'][frm,i,:,:]
        gsum[frm] = np.sum(hst['psino'][frm,:,:,:])
        gpu_totsum += gsum[frm]
        print('GPU('+str(frm)+') =', gsum[frm])
        print('-----------')
    print('GPUtot =', gpu_totsum)

    #---additional constants
    saggital_angle = 127
    coronal_angle = 0
    i_mxfrm = gsum.argmax()
    frmrep = 5
    mfrm = frmrep*nfrm
    #---

    #--for movie
    FFMpegWriter = manimation.writers['ffmpeg']
    metadata = dict(title='Axial View', artist='Pawel', comment='--')
    writer = FFMpegWriter(fps=10, bitrate=30000, metadata=metadata)
    #--

    fig1 = plt.figure()

    ax1 = plt.subplot(311)
    plt.title('Coronal View')
    plt.setp( ax1.get_xticklabels(), visible=False)
    plt.tick_params(axis='both', which='both', bottom='off', top='off', labelbottom='off')
    l1 = plt.imshow(np.array(ddsino[i_mxfrm, : , coronal_angle, :], dtype=np.float64), cmap='jet',interpolation='nearest')
    #plt.clim([0, 70])

    ax2 = plt.subplot(312)
    plt.title('Sagittal View')
    plt.setp( ax2.get_xticklabels(), visible=False)
    plt.tick_params(axis='both', which='both', bottom='off', top='off', labelbottom='off')
    l2 = plt.imshow(np.array(ddsino[i_mxfrm, : , saggital_angle, :], dtype=np.float64), cmap='jet',interpolation='nearest')
    #plt.clim([0, 70])

    ax3 = plt.subplot(313)
    plt.title('Axial Centre of Mass')
    plt.plot(range(hst['dur']), voxz*mvavg(hst['cmass'][:],5),'k')
    plt.ylim([voxz*ymin, voxz*ymax])
    plt.xlabel('Time [s]')
    l3, = plt.plot(np.array([1000, 1000]), np.array([0, ymax]), 'b')

    fnm = os.path.join(outpth, 'pViews_dyn.mp4')
    with writer.saving(fig1, fnm, 100):
        for frm in range(mfrm):
            print ('i> dynamic frame:', frm%nfrm)
            tmp = np.array(ddsino[frm%nfrm, : , coronal_angle, :], dtype=np.float64)
            l1.set_data(tmp)
            tmp = np.array(ddsino[frm%nfrm, : , saggital_angle, :], dtype=np.float64)
            l2.set_data(tmp)
            l3.set_data(frmcum[frm%nfrm]*np.ones(2), np.array([0, ymax]))
            writer.grab_frame()

    return fnm

#===================================================================================